import pp
from pp.components import bend_circular
from pp.components.waveguide import _arbitrary_straight_waveguide
from pp.components.waveguide import waveguide
from pp.components.waveguide_heater import waveguide_heater
from pp.components.taper import taper_strip_to_ridge

from pp import LAYER

from pp.components.component_sequence import component_sequence


@pp.autoname
def phase_modulator_waveguide(length, wg_width=0.5, cladding=3.0, si_outer_clad=1.0):
    """
    Phase modulator waveguide mockup
    """
    a = wg_width / 2
    b = a + cladding
    c = b + si_outer_clad

    windows = [
        (-c, -b, LAYER.WG),
        (-b, -a, LAYER.SLAB90),
        (-a, a, LAYER.WG),
        (a, b, LAYER.SLAB90),
        (b, c, LAYER.WG),
    ]

    component = _arbitrary_straight_waveguide(length=length, windows=windows)
    return component


@pp.autoname
def test_cutback_phase(straight_length=100.0, bend_radius=10.0, n=2):
    """ Modulator sections connected by bends """
    # Define sub components
    bend180 = bend_circular(radius=bend_radius, start_angle=-90, theta=180)
    pm_wg = phase_modulator_waveguide(length=straight_length)
    wg_short = waveguide(length=1.0)
    wg_short2 = waveguide(length=2.0)
    wg_heater = waveguide_heater(length=10.0)
    taper = taper_strip_to_ridge()

    # Define a map between symbols and (component, input port, output port)
    string_to_device_in_out_ports = {
        "I": (taper, "1", "wg_2"),
        "O": (taper, "wg_2", "1"),
        "S": (wg_short, "W0", "E0"),
        "P": (pm_wg, "W0", "E0"),
        "A": (bend180, "W0", "W1"),
        "B": (bend180, "W1", "W0"),
        "H": (wg_heater, "W0", "E0"),
        "-": (wg_short2, "W0", "E0"),
    }

    # Generate a sequence
    # This is simply a chain of characters. Each of them represents a component
    # with a given input and and a given output

    repeated_sequence = "SIPOSASIPOSB"
    heater_seq = "-H-H-H-H-"
    sequence = repeated_sequence * n + "SIPO" + heater_seq
    component = component_sequence(sequence, string_to_device_in_out_ports)

    assert component
    return component


if __name__ == "__main__":
    c = test_cutback_phase()
    pp.show(c)



"""
Connecting a component with I/O
"""

import numpy as np
import pp
from pp import LAYER
from pp import Port
from pp.routing.add_fiber_array import add_fiber_array


@pp.autoname
def big_device(w=400.0, h=400.0, N=16, port_pitch=15.0, layer=LAYER.WG, wg_width=0.5):
    """ big device with N ports on each side """
    component = pp.Component()
    p0 = np.array((0, 0))
    dx = w / 2
    dy = h / 2

    points = [[dx, dy], [dx, -dy], [-dx, -dy], [-dx, dy]]
    component.add_polygon(points, layer=layer)
    port_params = {"layer": layer, "width": wg_width}
    for i in range(N):
        port = Port(
            name="W{}".format(i),
            midpoint=p0 + (-dx, (i - N / 2) * port_pitch),
            orientation=180,
            **port_params
        )
        component.add_port(port)

    for i in range(N):
        port = Port(
            name="E{}".format(i),
            midpoint=p0 + (dx, (i - N / 2) * port_pitch),
            orientation=0,
            **port_params
        )
        component.add_port(port)

    for i in range(N):
        port = Port(
            name="N{}".format(i),
            midpoint=p0 + ((i - N / 2) * port_pitch, dy),
            orientation=90,
            **port_params
        )
        component.add_port(port)

    for i in range(N):
        port = Port(
            name="S{}".format(i),
            midpoint=p0 + ((i - N / 2) * port_pitch, -dy),
            orientation=-90,
            **port_params
        )
        component.add_port(port)
    return component


def test_big_device():
    component = big_device(N=10)
    bend_radius = 5.0
    c = add_fiber_array(component, bend_radius=bend_radius, fanout_length=50.0)
    assert c
    return c


if __name__ == "__main__":
    c = test_big_device()

    pp.show(c)



from typing import Any, List, Optional, Tuple
import pp
from pp.config import TAPER_LENGTH
from pp.component import Component


@pp.autoname
def taper(
    length: float = TAPER_LENGTH,
    width1: float = 0.5,
    width2: Optional[float] = None,
    port: None = None,
    layer: Tuple[int, int] = pp.LAYER.WG,
    layers_cladding: List[Any] = [],
    cladding_offset: float = 3.0,
) -> Component:
    """ Linear taper

    Args:
        length:
        width1:
        width2:
        port:
        layer:

    .. plot::
      :include-source:

      import pp

      c = pp.c.taper(width1=4, width2=1)
      pp.plotgds(c)

    """
    if type(port) is pp.Port and width1 is None:
        width1 = port.width
    if width2 is None:
        width2 = width1

    y1 = width1 / 2
    y2 = width2 / 2

    xpts = [0, length, length, 0]
    ypts = [y1, y2, -y2, -y1]

    c = pp.Component()
    c.add_polygon((xpts, ypts), layer=layer)
    c.add_port(name="1", midpoint=[0, 0], width=width1, orientation=180, layer=layer)
    c.add_port(name="2", midpoint=[length, 0], width=width2, orientation=0, layer=layer)

    o = cladding_offset
    ypts = [y1 + o, y2 + o, -y2 - o, -y1 - o]

    for layer in layers_cladding:
        c.add_polygon((xpts, ypts), layer=layer)

    c.info["length"] = length
    c.info["width1"] = width1
    c.info["width2"] = width2
    return c


@pp.autoname
def taper_strip_to_ridge(
    length: float = 10.0,
    width1: float = 0.5,
    width2: float = 0.5,
    w_slab1: float = 0.15,
    w_slab2: float = 5.0,
) -> Component:
    """ taper strip to rib

    Args:
        length:
        width1:
        width2:
        w_slab1
        w_slab2

    .. plot::
      :include-source:

      import pp

      c = pp.c.taper_strip_to_ridge()
      pp.plotgds(c)

    """

    _taper_wg = taper(length=length, width1=width1, width2=width2, layer=pp.LAYER.WG)
    _taper_slab = taper(
        length=length, width1=w_slab1, width2=w_slab2, layer=pp.LAYER.SLAB90
    )

    c = pp.Component()
    for _t in [_taper_wg, _taper_slab]:
        taper_ref = _t.ref()
        c.add(taper_ref)
        c.absorb(taper_ref)

    c.info["length"] = length
    c.add_port(name="1", port=_taper_wg.ports["1"])
    c.add_port(name="wg_2", port=_taper_wg.ports["2"])
    c.add_port(name="slab_2", port=_taper_slab.ports["2"])

    return c


@pp.autoname
def taper_strip_to_ridge_trenches(
    length=10.0,
    width=0.5,
    slab_offset=3.0,
    trench_width=2.0,
    trench_layer=pp.LAYER.SLAB90,
    wg_layer=pp.LAYER.WG,
    trench_offset_after_wg=0.1,
):

    c = pp.Component()
    width = pp.bias.width(width)
    y0 = width / 2 + trench_width - trench_offset_after_wg
    yL = width / 2 + trench_width - trench_offset_after_wg + slab_offset

    # waveguide
    x = [0, length, length, 0]
    yw = [y0, yL, -yL, -y0]
    c.add_polygon((x, yw), layer=wg_layer)

    # top trench
    ymin0 = width / 2
    yminL = width / 2
    ymax0 = width / 2 + trench_width
    ymaxL = width / 2 + trench_width + slab_offset
    x = [0, length, length, 0]
    ytt = [ymin0, yminL, ymaxL, ymax0]
    ytb = [-ymin0, -yminL, -ymaxL, -ymax0]
    c.add_polygon((x, ytt), layer=trench_layer)
    c.add_polygon((x, ytb), layer=trench_layer)

    c.add_port(name="W0", midpoint=[0, 0], width=width, orientation=180, layer=wg_layer)
    c.add_port(
        name="E0", midpoint=[length, 0], width=width, orientation=0, layer=wg_layer
    )

    return c


if __name__ == "__main__":
    c = taper(width2=1, layers_cladding=[pp.LAYER.WGCLAD])
    # c = taper_strip_to_ridge()
    # print(c.get_optical_ports())
    # c = taper_strip_to_ridge_trenches()
    pp.show(c)



"""
# Components

We store our component functions inside the `pp.components` module. Each function there returns a Component object

You can use `dir` the `pp.components` module to see the all available components.

Some of which are just shapes, but we call them components as they all inherit from the component class in `pp.Component`
"""


import pp


def test_components_module():
    assert len(dir(pp.c)) > 1


if __name__ == "__main__":
    print(dir(pp.c))
    print(len(dir(pp.c)))



""" based on phidl tutorial

# Connecting devices with connect()

The connect command allows us to connect DeviceReference ports together like
Lego blocks.  There is an optional parameter called ``overlap`` which is
useful if you have shapes you want to intersect with some overlap (or with a
negative number, separate the ports).

"""


import pp

c = pp.Component("waveguides_sample")


wg1 = c << pp.c.waveguide(length=10, width=1)
wg2 = c << pp.c.waveguide(length=10, width=2, layer=pp.LAYER.SLAB90)
wg3 = c << pp.c.waveguide(length=10, width=3, layer=pp.LAYER.SLAB150)


wg2.connect(port="W0", destination=wg1.ports["E0"])
wg3.connect(port="W0", destination=wg2.ports["E0"], overlap=1)

pp.qp(c)  # quickplot it!
pp.show(c)  # show it in klayout



from pp.container import container
from pp.component import Component
from pp.components.electrical.pad import pad
from pp.routing.connect_electrical import connect_electrical_shortest_path


@container
def add_electrical_pads_shortest(component, pad=pad, pad_port_spacing=50, **kwargs):
    """add a pad to each closest electrical port
    Args:
        component:
        pad: pad element or function
        pad_port_spacing: between pad and port
        width: pad width
        height: pad height
        layer: pad layer

    """
    c = Component(f"{component}_e")
    ports = component.get_ports_list(port_type="dc")
    c << component

    pad = pad(**kwargs) if callable(pad) else pad
    pad_port_spacing += pad.settings["width"] / 2

    for port in ports:
        p = c << pad
        if port.orientation == 0:
            p.x = port.x + pad_port_spacing
            p.y = port.y
            c.add(connect_electrical_shortest_path(port, p.ports["W"]))
        elif port.orientation == 180:
            p.x = port.x - pad_port_spacing
            p.y = port.y
            c.add(connect_electrical_shortest_path(port, p.ports["E"]))
        elif port.orientation == 90:
            p.y = port.y + pad_port_spacing
            p.x = port.x
            c.add(connect_electrical_shortest_path(port, p.ports["S"]))
        elif port.orientation == 270:
            p.y = port.y - pad_port_spacing
            p.x = port.x
            c.add(connect_electrical_shortest_path(port, p.ports["N"]))

    c.ports = component.ports
    for port in ports:
        c.ports.pop(port.name)
    return c


if __name__ == "__main__":
    import pp

    c = pp.c.cross(length=100, layer=pp.LAYER.M3, port_type="dc")
    c.move((20, 50))
    cc = add_electrical_pads_shortest(c)

    pp.show(cc)



from typing import List, Tuple

import hashlib
import pp
from pp.name import autoname
from pp.components.hline import hline

from pp.component import Component


@autoname
def waveguide(
    length: float = 10.0,
    width: float = 0.5,
    layer: Tuple[int, int] = pp.LAYER.WG,
    layers_cladding: List[Tuple[int, int]] = [pp.LAYER.WGCLAD],
    cladding_offset: float = 3.0,
) -> Component:
    """ straight waveguide

    Args:
        length: in X direction
        width: in Y direction

    .. plot::
      :include-source:

      import pp

      c = pp.c.waveguide(length=10, width=0.5)
      pp.plotgds(c)

    """
    c = Component()
    w = width / 2
    c.add_polygon([(0, -w), (length, -w), (length, w), (0, w)], layer=layer)

    wc = w + cladding_offset

    for layer_cladding in layers_cladding:
        c.add_polygon(
            [(0, -wc), (length, -wc), (length, wc), (0, wc)], layer=layer_cladding
        )

    c.add_port(name="W0", midpoint=[0, 0], width=width, orientation=180, layer=layer)
    c.add_port(name="E0", midpoint=[length, 0], width=width, orientation=0, layer=layer)

    c.width = width
    c.length = length
    return c


@autoname
def wg_shallow_rib(width=0.5, layer=pp.LAYER.SLAB150, layers_cladding=[], **kwargs):
    width = pp.bias.width(width)
    return waveguide(
        width=width, layer=layer, layers_cladding=layers_cladding, **kwargs
    )


@autoname
def wg_deep_rib(width=0.5, layer=pp.LAYER.SLAB90, layers_cladding=[], **kwargs):
    width = pp.bias.width(width)
    return waveguide(
        width=width, layer=layer, layers_cladding=layers_cladding, **kwargs
    )


@autoname
def waveguide_biased(width=0.5, **kwargs):
    width = pp.bias.width(width)
    return waveguide(width=width, **kwargs)


def _arbitrary_straight_waveguide(length, windows):
    """
    windows: [(y_start, y_stop, layer), ...]
    """
    md5 = hashlib.md5()
    for e in windows:
        md5.update(str(e).encode())

    component = Component()
    component.name = "ARB_SW_L{}_HASH{}".format(length, md5.hexdigest())
    y_min, y_max, layer0 = windows[0]
    y_min, y_max = min(y_min, y_max), max(y_min, y_max)

    # Add one port on each side centered at y=0
    for y_start, y_stop, layer in windows:
        w = abs(y_stop - y_start)
        y = (y_stop + y_start) / 2
        _wg = hline(length=length, width=w, layer=layer).ref()
        _wg.movey(y)
        component.add(_wg)
        component.absorb(_wg)
        y_min = min(y_stop, y_start, y_min)
        y_max = max(y_stop, y_start, y_max)
    width = y_max - y_min

    component.add_port(
        name="W0", midpoint=[0, 0], width=width, orientation=180, layer=layer0
    )
    component.add_port(
        name="E0", midpoint=[length, 0], width=width, orientation=0, layer=layer0
    )

    return component


@autoname
def waveguide_slab(length=10.0, width=0.5, cladding=2.0, slab_layer=pp.LAYER.SLAB90):
    width = pp.bias.width(width)
    ymin = width / 2
    ymax = ymin + cladding
    windows = [(-ymin, ymin, pp.LAYER.WG), (-ymax, ymax, slab_layer)]
    return _arbitrary_straight_waveguide(length=length, windows=windows)


@autoname
def waveguide_trenches(
    length=10.0,
    width=0.5,
    layer=pp.LAYER.WG,
    trench_width=3.0,
    trench_offset=0.2,
    trench_layer=pp.LAYER.SLAB90,
):
    width = pp.bias.width(width)
    w = width / 2
    ww = w + trench_width
    wt = ww + trench_offset
    windows = [(-ww, ww, layer), (-wt, -w, trench_layer), (w, wt, trench_layer)]
    return _arbitrary_straight_waveguide(length=length, windows=windows)


waveguide_ridge = waveguide_slab


@autoname
def waveguide_slot(length=10.0, width=0.5, gap=0.2, layer=pp.LAYER.WG):
    width = pp.bias.width(width)
    gap = pp.bias.gap(gap)
    a = width / 2
    d = a + gap / 2

    windows = [(-a - d, a - d, layer), (-a + d, a + d, layer)]
    return _arbitrary_straight_waveguide(length=length, windows=windows)


def _demo_waveguide():
    c = waveguide()
    pp.write_gds(c)
    return c


if __name__ == "__main__":

    c = waveguide(length=4, pins=True)

    pp.show(c)

    # print(c.hash_geometry())
    # pp.show(c)

    # print(c.ports)
    # cc = pp.routing.add_fiber_array(c)
    # pp.show(cc)

    # c = waveguide_slab()
    # c = waveguide_trenches()
    # c = waveguide()
    # c = waveguide_slot()
    # c = waveguide_slot(length=11.2, width=0.5)
    # c = waveguide_slot(length=11.2, width=0.5)
    # pp.show(c)



from typing import Callable, Optional, Tuple
import picwriter.components as pc
from pp.port import auto_rename_ports

from pp.components.waveguide_template import wg_strip
from pp.picwriter2component import picwriter2component
import pp
from pp.component import Component


@pp.autoname
def spiral(
    width: float = 500.0,
    length: float = 10e3,
    spacing: Optional[float] = None,
    parity: int = 1,
    port: Tuple[int, int] = (0, 0),
    direction: str = "NORTH",
    waveguide_template: Callable = wg_strip,
    **kwargs
) -> Component:
    """ Picwriter Spiral

    Args:
       width (float): width of the spiral (i.e. distance between input/output ports)
       length (float): desired length of the waveguide (um)
       spacing (float): distance between parallel waveguides
       parity (int): If 1 spiral on right side, if -1 spiral on left side (mirror flip)
       port (tuple): Cartesian coordinate of the input port
       direction (string): Direction that the component will point *towards*, can be of type `'NORTH'`, `'WEST'`, `'SOUTH'`, `'EAST'`, OR an angle (float, in radians)
       waveguide_template (WaveguideTemplate): Picwriter WaveguideTemplate object
       wg_width: 0.5
       wg_layer: pp.LAYER.WG[0]
       wg_datatype: pp.LAYER.WG[1]
       clad_layer: pp.LAYER.WGCLAD[0]
       clad_datatype: pp.LAYER.WGCLAD[1]
       bend_radius: 10
       cladding_offset: 3

    .. plot::
      :include-source:

      import pp

      c = pp.c.spiral(width=500, length=10e3)
      pp.plotgds(c)

    """
    c = pc.Spiral(
        pp.call_if_func(waveguide_template, **kwargs),
        width=width,
        length=length,
        spacing=spacing,
        parity=parity,
        port=port,
        direction=direction,
    )
    # print(f'length = {length/1e4:.2f}cm')
    c = picwriter2component(c)
    c = auto_rename_ports(c)
    return c


if __name__ == "__main__":
    c = spiral(length=10e3, width=500, bend_radius=20, pins=True)
    pp.show(c)



"""
Builds a library and compares the gdshash of the new built GDS with the reference ones
"""

import os
import shutil
import pytest
from jsondiff import diff
import git

import gdspy
import pp
from pp.components import component_type2factory, _components
from pp import CONFIG

path_library = CONFIG["gds"]

_components = _components - set(
    [
        "compass",
        "component_lattice",
        "grating_coupler_elliptical2",
        "grating_coupler_tree",
        "bezier",
        "spiral_circular",
        "coupler90",
        "waveguide_pin",
        "waveguide_heater",
        "wg_heater_connected",
    ]
)


def pull_library(path_library=path_library):
    try:
        if os.path.isdir(path_library):
            print("git pull: {}".format(path_library))
            g = git.cmd.Git(path_library)
            g.pull()
    except Exception as e:
        error = f"error pulling {path_library} repo"
        raise ValueError(error, e)


def lock_component(
    component_type,
    component_type2factory=component_type2factory,
    path_library=path_library,
    flatten=True,
    with_settings_label=False,
):
    """ locks a component from the factory into the GDS lib

    TODO: flattening makes it slower for big mask
    """
    try:
        c = component_type2factory[component_type](cache=False, pins=False)
        if flatten:
            c.flatten()
        gdspath = path_library / (component_type + ".gds")
        pp.write_component(
            c, gdspath=gdspath, with_settings_label=with_settings_label,
        )
        assert gdspath.exists()
        return c
    except Exception as e:
        error = f"error building {component_type}"
        raise ValueError(error, e)


def lock_components_with_changes(
    components=_components,
    component_type2factory=component_type2factory,
    path_library=path_library,
):

    """  locks only components whose hash changed
    """
    for component_type in components:
        same_hash = compare_component_hash(
            component_type=component_type,
            component_type2factory=component_type2factory,
            path_library=path_library,
        )
        if not same_hash:
            print(f"locking {component_type} to {path_library}")
            lock_component(
                component_type=component_type,
                component_type2factory=component_type2factory,
                path_library=path_library,
            )


def print_components_with_changes(
    component_type2factory=component_type2factory, path_library=path_library
):
    """  locks only components whose hash changed
    """
    for component_type, _ in component_type2factory.items():
        same_hash = compare_component_hash(
            component_type,
            component_type2factory=component_type2factory,
            path_library=path_library,
        )
        if not same_hash:
            print(f"[X] {component_type} changed hash")


@pytest.mark.parametrize("component_type", _components)
@pytest.mark.noautofixt
def test_components(
    component_type,
    component_type2factory=component_type2factory,
    path_library=path_library,
):
    assert compare_component_hash(
        component_type=component_type,
        component_type2factory=component_type2factory,
        path_library=path_library,
    ), f"{component_type} changed from component locked in the library {path_library}"


def rebuild_library(
    component_type2factory=component_type2factory, path_library=path_library
):
    """ saves all component_type2factory components to the gdslib library
    """
    for component_type, _ in component_type2factory.items():
        lock_component(
            component_type=component_type,
            component_type2factory=component_type2factory,
            path_library=path_library,
        )
        print(f"Write `{component_type}`")


def _copy_component(src, dest):
    for ext in [".gds", ".json", ".ports"]:
        shutil.copy(src.with_suffix(ext), dest.with_suffix(ext))

    for ext in [".ports"]:
        try:
            shutil.copy(src.with_suffix(ext), dest.with_suffix(ext))
        except:
            pass


def compare_component_hash(
    component_type,
    component_type2factory=component_type2factory,
    path_library=path_library,
    path_test=CONFIG["gdslib_test"],
):
    """ raises Exception if component has changed from the library
    writes component if it does not exist

    Args:
        component_type:
        path_library:
        path_test:
    """

    component_new = lock_component(
        component_type,
        component_type2factory=component_type2factory,
        path_library=path_test,
    )
    component_new.name += "_new"

    gdspath_new = path_test / (component_type + ".gds")
    gdspath_library = path_library / (component_type + ".gds")

    if not gdspath_library.exists():
        lock_component(
            component_type=component_type,
            component_type2factory=component_type2factory,
            path_library=path_library,
        )
        print(f"writing new component {component_type} into {path_library}")
        return True

    component_library = pp.load_component(
        component_name=component_type, component_path=path_library
    )
    component_library.name += "_lib"
    gdshash_new = gdspy.gdsii_hash(gdspath_new)
    gdshash_library = gdspy.gdsii_hash(gdspath_library)

    # gdshash_new = component_library.hash_geometry()
    # gdshash_library = component_new.hash_geometry()

    same_hash = gdshash_new == gdshash_library
    if not same_hash:
        error_hash = f"`{component_library}` hash(GDS) {gdspath_new} differs from the library {gdspath_library}, showing both cells in Klayout \n library = {component_type2factory.keys()}"
        error_settings = f"different settings: {diff(component_library.get_settings(), component_new.get_settings())}"
        c = pp.Component(name=component_type)
        c << component_new
        c << component_library
        pp.show(c)
        print(error_hash + error_settings)

    return same_hash


if __name__ == "__main__":
    # for component in list(_components):
    #     test_components(component)
    lock_components_with_changes()
    # lock_component("grating_coupler_tree")
    # compare_component_hash("grating_coupler_tree")
    # rebuild_library()
    # lock_component("waveguide")
    # compare_component_hash("waveguide")
    # lock_component("ring_double_bus")
    # compare_component_hash("ring_double_bus")
    # compare_component_hash("coupler90")
    # print_components_with_changes()



import numpy as np
import pp
from pp.geo_utils import DEG2RAD
from pp.components.grating_coupler.elliptical import grating_tooth_points
from pp.component import Component
from typing import Tuple


@pp.autoname
def grating_coupler_elliptical_trenches(
    polarization: str = "te",
    taper_length: float = 16.6,
    taper_angle: float = 30.0,
    trenches_extra_angle: float = 9.0,
    lambda_c: float = 1.53,
    fiber_angle: float = 15.0,
    grating_line_width: float = 0.343,
    wg_width: float = 0.5,
    neff: float = 2.638,  # tooth effective index
    layer: Tuple[int, int] = pp.LAYER.WG,
    layer_trench: Tuple[int, int] = pp.LAYER.SLAB150,
    p_start: int = 26,
    n_periods: int = 30,
    straight: float = 0.2,
) -> Component:
    """ Grating coupler

    Args:
        polarization: 'te' or 'tm'
        taper_length: taper length from waveguide I/O
        taper_angle: grating flare angle
        lambda_c: grating transmission central wavelength
        fiber_angle: fibre polish angle in degrees
        grating_line_width: of the 220 ridge
        wg_width: waveguide width
        neff: 2.638  # tooth effective index
        layer: LAYER.WG
        layer_trench: LAYER.SLAB150
        p_start: 26  # first tooth
        n_periods: 16  # number of periods
        straight: 0.2

    .. plot::
      :include-source:

      import pp
      from pp.components.grating_coupler.elliptical_trenches import grating_coupler_elliptical_trenches

      c = grating_coupler_elliptical_trenches()
      pp.plotgds(c)
    """

    # Define some constants
    nc = 1.443  # cladding index

    # Compute some ellipse parameters
    sthc = np.sin(fiber_angle * DEG2RAD)
    d = neff ** 2 - nc ** 2 * sthc ** 2
    a1 = lambda_c * neff / d
    b1 = lambda_c / np.sqrt(d)
    x1 = lambda_c * nc * sthc / d

    a1 = round(a1, 3)
    b1 = round(b1, 3)
    x1 = round(x1, 3)

    period = a1 + x1
    trench_line_width = period - grating_line_width

    c = pp.Component()
    c.polarization = polarization
    c.wavelength = int(lambda_c * 1e3)

    # Make each grating line
    for p in range(p_start, p_start + n_periods + 1):
        pts = grating_tooth_points(
            p * a1,
            p * b1,
            p * x1,
            width=trench_line_width,
            taper_angle=taper_angle + trenches_extra_angle,
        )
        c.add_polygon(pts, layer_trench)

    # Make the taper
    p_taper = p_start - 1
    p_taper_eff = p_taper
    a_taper = a1 * p_taper_eff
    # b_taper = b1 * p_taper_eff
    x_taper = x1 * p_taper_eff
    x_output = a_taper + x_taper - taper_length + grating_line_width / 2

    xmax = x_output + taper_length + n_periods * period + 3
    y = wg_width / 2 + np.tan(taper_angle / 2 * np.pi / 180) * xmax
    pts = [
        (x_output, -wg_width / 2),
        (x_output, wg_width / 2),
        (xmax, y),
        (xmax + straight, y),
        (xmax + straight, -y),
        (xmax, -y),
    ]
    c.add_polygon(pts, layer)

    # a_taper = (p_start + n_periods + 1)*a1
    # b_taper = (p_start + n_periods + 1)*b1
    # pts = grating_taper_points(
    #     a_taper, b_taper, x_output, x_taper, taper_angle, wg_width=wg_width
    # )
    # c.add_polygon(pts, layer)

    # Move waveguide I/O to (0, 0)
    c.move((-x_output, 0))

    if polarization.lower() == "te":
        polarization_marker_layer = pp.LAYER.polarization_te
    else:
        polarization_marker_layer = pp.LAYER.polarization_tm

    circle = pp.c.circle(radius=17 / 2, layer=polarization_marker_layer)
    circle_ref = c.add_ref(circle)
    circle_ref.movex(taper_length + period * n_periods / 2)

    # Add port
    c.add_port(name="W0", midpoint=[0, 0], width=wg_width, orientation=180, layer=layer)
    c.settings["period"] = period
    return c


def grating_coupler_te(taper_angle: int = 35, **kwargs) -> Component:
    """

    .. plot::
      :include-source:

      import pp

      c = pp.c.grating_coupler_te()
      pp.plotgds(c)
    """
    return grating_coupler_elliptical_trenches(
        polarization="te", taper_angle=taper_angle, **kwargs
    )


def grating_coupler_tm(
    neff: float = 1.8, grating_line_width: float = 0.6, **kwargs
) -> Component:
    """

    .. plot::
      :include-source:

      import pp

      c = pp.c.grating_coupler_tm()
      pp.plotgds(c)
    """
    return grating_coupler_elliptical_trenches(
        polarization="tm", neff=neff, grating_line_width=grating_line_width, **kwargs
    )


if __name__ == "__main__":
    # c = grating_coupler_elliptical_trenches(polarization="TE")
    # print(c.polarization)
    c = grating_coupler_te()
    # c = grating_coupler_tm()
    pp.show(c)



""" gdspy based placer

YAML defines component DOE settings and placement

.. code:: yaml

    placer:
        width: 30000  # um
        height: 20000 # um
        x_spacing: 1000 # x spacing origin to origin between the components
        y_spacing: 500 #  y spacing origin to origin between the components
        x_start: 500
        y_start: 100

    mmi1x2_gap:
        doe_name: doe1
        component: mmi1x2
        gap: [0.5, 0.6]
        length: 10

    mmi1x2:
        doe_name: doe1
        component: mmi1x2
        length: [11, 12]
        gap: [0.2, 0.3]
        do_permutation: False

    placement:
        A-B1-2: doe1
"""

import os
import sys
from omegaconf import OmegaConf

import pp
from pp.doe import get_settings_list, load_does
from pp.config import CONFIG
from pp.components import component_type2factory
from pp.write_component import write_gds
from pp.write_component import write_component_report


def _print(*args, **kwargs):
    print(*args, **kwargs)
    sys.stdout.flush()


def placer_grid_cell_refs(
    component_factory, cols=1, rows=1, dx=10.0, dy=10.0, x0=0, y0=0, **settings
):
    print(component_factory)
    if callable(component_factory):
        settings_list = get_settings_list(**settings)
        component_list = [component_factory(**s) for s in settings_list]
    else:
        component_list = component_factory

    indices = [(i, j) for j in range(cols) for i in range(rows)]

    if rows * cols < len(component_list):
        raise ValueError(
            "Shape ({}, {}): Not enough emplacements ({}) for all these components ({}).".format(
                rows, cols, len(indices), len(component_list)
            )
        )
    components = []
    for component, (i, j) in zip(component_list, indices):
        c_ref = component.ref(position=(x0 + j * dx, y0 + i * dy))
        components += [c_ref]

    return components


def pack_horizontal(
    cells, row_ids=None, x0=0, y0=0, align_x="W", align_y="S", margin_x=20, margin_y=20,
):
    """
    Args:
        cells: a list of cells  (size n)
        row_ids: a list of row ids (size n)
            where each id represents the row where the cell should be placed
            None by default => all cells in the same row

    returns a list of cell references
    """
    heights = [c.size_info.height for c in cells]

    row_ids = row_ids or [0] * len(cells)

    if len(cells) != len(row_ids):
        raise ValueError(
            "Each cell should be assigned a row id. \
        Got {} cells for {} row ids".format(
                len(cells), len(row_ids)
            )
        )

    # Find the height of each row to fit the cells
    # Also group the cells by row

    unique_row_ids = list(set(row_ids))
    unique_row_ids.sort()
    _row_to_heights = {r: [] for r in set(row_ids)}
    row_to_cells = {r: [] for r in unique_row_ids}
    for row, h, cell in zip(row_ids, heights, cells):
        _row_to_heights[row] += [h]
        row_to_cells[row] += [cell]

    row_to_height = {k: max(v) for k, v in _row_to_heights.items()}

    components = []

    # Do the packing per row
    y = y0
    for row in unique_row_ids:
        cells = row_to_cells[row]
        x = x0

        for c in cells:
            if align_x == "W" and align_y == "S":
                component_origin = c.size_info.sw
            elif align_x == "E" and align_y == "S":
                component_origin = c.size_info.se
            elif align_x == "E" and align_y == "N":
                component_origin = c.size_info.ne
            elif align_x == "W" and align_y == "N":
                component_origin = c.size_info.nw
            try:
                components += [c.ref(position=-component_origin + (x, y))]
            except:
                if align_x not in ["W", "E"]:
                    print("align_x should be `W`, `E` or a float")
                if align_y not in ["N", "S"]:
                    print("align_y should be `N`, `S` or a float")
                raise

            if align_x == "W":
                x += c.size_info.width + margin_x
            else:
                x += -c.size_info.width - margin_x

        if align_y == "S":
            y += row_to_height[row] + margin_y
        else:
            y += -row_to_height[row] - margin_y

    return components


def pack_vertical(
    cells, col_ids=None, x0=0, y0=0, align_x="W", align_y="S", margin_x=20, margin_y=20,
):
    """
    Args:
        cells: a list of cells  (size n)
        col_ids: a list of column ids (size n)
            where each id represents the row where the cell should be placed
            None by default => all cells are packed in the same column

    returns a list of cell references
    """
    widths = [c.size_info.width for c in cells]
    col_ids = col_ids or [0] * len(cells)

    if len(cells) != len(col_ids):
        raise ValueError(
            "Each cell should be assigned a row id. \
        Got {} cells for {} col ids".format(
                len(cells), len(col_ids)
            )
        )

    # Find the width of each column to fit the cells
    # Also group the cells by column

    unique_col_ids = list(set(col_ids))
    unique_col_ids.sort()
    _col_to_widths = {r: [] for r in set(col_ids)}
    col_to_cells = {r: [] for r in unique_col_ids}
    for col, w, cell in zip(col_ids, widths, cells):
        _col_to_widths[col] += [w]
        col_to_cells[col] += [cell]

    col_to_width = {k: max(v) for k, v in _col_to_widths.items()}

    components = []

    # Do the packing per column
    x = x0
    for col in unique_col_ids:
        cells = col_to_cells[col]
        y = y0

        for c in cells:
            if align_x == "W" and align_y == "S":
                component_origin = c.size_info.sw
            elif align_x == "E" and align_y == "S":
                component_origin = c.size_info.se
            elif align_x == "E" and align_y == "N":
                component_origin = c.size_info.ne
            elif align_x == "W" and align_y == "N":
                component_origin = c.size_info.nw
            components += [c.ref(position=-component_origin + (x, y))]

            if align_y == "S":
                y += c.size_info.height + margin_y
            else:
                y += -c.size_info.height - margin_y

        if align_x == "W":
            x += col_to_width[col] + margin_x
        else:
            x += -col_to_width[col] - margin_x

    return components


def placer_fixed_coords(cells, x, y, **kwargs):
    if len(cells) == 1:
        cells = cells * len(x)

    return [c.ref(position=(_x, _y)) for c, _x, _y in zip(cells, x, y)]


PLACER_NAME2FUNC = {
    "grid": placer_grid_cell_refs,
    "pack_row": pack_horizontal,
    "pack_col": pack_vertical,
    "fixed_coords": placer_fixed_coords,
}


def load_placer_with_does(filepath, defaults={"do_permutation": True}):
    """ load placer settings

    Args:
        filepath: a yaml file containing the does and placer information

    Returns:
        a dictionnary of DOEs with:
        {
            doe_name1: [(component_factory_name, parameters), ...]
            doe_name2: [(component_factory_name, parameters), ...]
            ...
        }

    """
    does = {}
    data = OmegaConf.load(filepath)

    placer_info = data.pop("placer")
    component_placement = data.pop("placement")
    gds_files = {}
    if "gds" in data.keys():
        gds_files.update(data.pop("gds"))

    for doe_chunk in data.values():
        doe_name = doe_chunk.pop("doe_name")
        component_type = doe_chunk.pop("component")

        if doe_name not in does:
            does[doe_name] = {"settings": [], "component_type": component_type}

        do_permutation = defaults["do_permutation"]
        if "do_permutation" in doe_chunk:
            do_permutation = doe_chunk.pop("do_permutation")

        doe = does[doe_name]
        # All the remaining parameters are component parameters
        if "settings" in doe_chunk:
            settings = doe_chunk.pop("settings")
        else:
            settings = {}
        doe["list_settings"] += get_settings_list(do_permutation, **settings)

        # check that the doe is valid (only one type of component)
        assert (
            component_type == doe["component_type"]
        ), "There can be only one component type per doe. Got {} while expecting {}".format(
            component_type, doe["component_type"]
        )

    return does, placer_info, component_placement, gds_files


CONTENT_SEP = " , "


def save_doe(doe_name, components, doe_root_path=None, precision=1e-9):
    """
    Save all components from this DOE in a tmp cache folder
    """
    if doe_root_path is None:
        doe_root_path = CONFIG["cache_doe_directory"]
    doe_dir = os.path.join(doe_root_path, doe_name)
    if not os.path.exists(doe_dir):
        os.makedirs(doe_dir)

    # Store list of component names - order matters
    component_names = [c.name for c in components]
    content_file = os.path.join(doe_dir, "content.txt")
    with open(content_file, "w") as fw:
        fw.write(CONTENT_SEP.join(component_names))

    for c in components:
        gdspath = os.path.join(doe_dir, c.name + ".gds")
        write_gds(c, gdspath=gdspath, precision=precision)
        write_component_report(c, json_path=gdspath[:-4] + ".json")


def load_doe_from_cache(doe_name, doe_root_path=None):
    """
    Load all components for this DOE from the cache
    """
    if doe_root_path is None:
        doe_root_path = CONFIG["cache_doe_directory"]
    doe_dir = os.path.join(doe_root_path, doe_name)
    content_file = os.path.join(doe_dir, "content.txt")
    with open(content_file) as f:
        component_names = f.read().split(CONTENT_SEP)

    gdspaths = [os.path.join(doe_dir, name + ".gds") for name in component_names]
    components = [pp.import_gds(gdspath) for gdspath in gdspaths]
    return components


def load_doe_component_names(doe_name, doe_root_path=None):
    if doe_root_path is None:
        doe_root_path = CONFIG["cache_doe_directory"]
    doe_dir = os.path.join(doe_root_path, doe_name)
    content_file = os.path.join(doe_dir, "content.txt")

    with open(content_file) as f:
        component_names = f.read().split(CONTENT_SEP)
    return component_names


def doe_exists(doe_name, list_settings, doe_root_path=None):
    """
    Check whether the folder exists and that the number of items in content.txt
    matches the number of items in list_settings
    """
    if doe_root_path is None:
        doe_root_path = CONFIG["cache_doe_directory"]
    doe_dir = os.path.join(doe_root_path, doe_name)
    content_file = os.path.join(doe_dir, "content.txt")
    if not os.path.exists(content_file):
        return False
    with open(content_file) as f:
        component_names = f.read().split(CONTENT_SEP)

    if len(component_names) == len(list_settings) or (
        len(list_settings) == 0 and len(component_names) == 1
    ):
        return True

    print(
        "doe_exists - DOE",
        doe_name,
        "needs regeneration",
        len(component_names),
        len(list_settings),
    )
    return False


def component_grid_from_yaml(filepath, precision=1e-9):
    """ Returns a Component composed of DOEs/components given in a yaml file
    allows for each DOE to have its own x and y spacing (more flexible than method1)
    """
    input_does = OmegaConf.load(str(filepath))
    mask_settings = input_does["mask"]
    does = load_does(filepath)

    placed_doe = None
    placed_does = {}
    if mask_settings.get("name"):
        component_grid = pp.Component(mask_settings["name"])
    else:
        component_grid = pp.Component()

    default_cache_enabled = (
        mask_settings["cache_enabled"] if "cache_enabled" in mask_settings else False
    )
    for doe_name, doe in does.items():
        list_settings = doe["settings"]
        component_type = doe["component"]

        # description = doe["description"] if "description" in doe else ''
        # test = doe["test"] if "test" in doe else {}
        # analysis = doe["analysis"] if "analysis" in doe else {}

        # Get DOE policy concerning the cache
        cache_enabled = (
            doe["cache_enabled"] if "cache_enabled" in doe else default_cache_enabled
        )

        components = None

        # If cache enabled, attempt to load from cache
        if cache_enabled:
            try:
                components = load_doe_from_cache(doe_name)
            except Exception as e:
                _print(e)
                components = None

        # If no component is loaded, build them
        if components is None:
            print("{} - Generating components...".format(doe_name))
            # If a generator is specified for the DOE, use it
            if "generator" in doe:
                components = _gen_components_from_generator(
                    doe["generator"], component_type, list_settings
                )
            # Otherwise generate each component using the component factory
            else:
                components = _gen_components(component_type, list_settings)

            # After building the components, if cache enabled, save them
            if cache_enabled:
                save_doe(doe_name, components, precision=precision)
        else:
            _print("{} - Loaded components from cache".format(doe_name))

        # _print(doe_name, [c.name for c in components])
        # Find placer information

        default_settings = {"align_x": "W", "align_y": "S", "margin": 10}

        if "placer" in doe:
            placer_type = doe["placer"].pop("type")
            _placer = PLACER_NAME2FUNC[placer_type]

            # All other attributes are assumed to be settings for the placer
            settings = default_settings.copy()
            settings.update(doe["placer"])

            # x0, y0 can either be float or string
            x0 = settings.pop("x0")
            y0 = settings.pop("y0")

            # Check whether we are doing relative or absolute placement
            if (x0 in ["E", "W"] or y0 in ["N", "S"]) and not placed_doe:
                raise ValueError(
                    "At least one DOE must be placed to use\
                relative placement"
                )

            # For relative placement (to previous DOE)
            if "margin_x" not in settings:
                settings["margin_x"] = settings["margin"]
            if "margin_y" not in settings:
                settings["margin_y"] = settings["margin"]

            margin_x = settings["margin_x"]
            margin_y = settings["margin_y"]
            align_x = settings["align_x"]
            align_y = settings["align_y"]

            ## Making sure that the alignment is sensible depending on how we stack

            # If we specify a DOE to place next to, use it
            if "next_to" in settings:
                placed_doe = placed_does[settings.pop("next_to")]

            # Otherwise, use previously placed DOE as starting point
            if x0 == "E":
                x0 = placed_doe.size_info.east
                if align_x == "W":
                    x0 += margin_x

            if x0 == "W":
                x0 = placed_doe.size_info.west
                if align_x == "E":
                    x0 -= margin_x

            if y0 == "N":
                y0 = placed_doe.size_info.north
                if align_y == "S":
                    y0 += margin_y

            if y0 == "S":
                y0 = placed_doe.size_info.south
                if align_y == "N":
                    y0 -= margin_y

            # Add x0, y0 in settings as float
            settings["x0"] = x0
            settings["y0"] = y0

            placed_components = _placer(components, **settings)
        else:
            # If no placer is specified, we assume this is a grid placer
            cols, rows = doe["shape"]
            x0, y0 = doe["origin"]
            dx, dy = doe["spacing"]

            placed_components = placer_grid_cell_refs(
                components, cols, rows, dx, dy, x0=x0, y0=y0
            )

        # Place components within a cell having the DOE name
        placed_doe = pp.Component()
        placed_doe.add(placed_components)
        placed_doe.name = doe_name
        placed_does[doe_name] = placed_doe

        # # Write the json and md metadata / report
        # write_doe_metadata(
        # doe_name=doe_name,
        # cells = placed_components,
        # list_settings=list_settings,
        # flag_write_component=False,
        # description=description,
        # test=test,
        # analysis=analysis
        # )

        component_grid.add_ref(placed_doe)

    return component_grid


def _gen_components(
    component_type, list_settings, component_type2factory=component_type2factory
):
    components = []

    # If no settings passed, generate a single component with defaults
    if not list_settings:
        component_factory = component_type2factory[component_type]
        component = component_factory()
        component.function_name = component_factory.__name__
        components += [component]
        return components

    for settings in list_settings:
        component_factory = component_type2factory[component_type]
        component = component_factory(**settings)
        component.function_name = component_factory.__name__
        components += [component]

    return components


def _gen_components_from_generator(
    components_generator,
    component_type,
    doe,
    component_type2factory=component_type2factory,
):
    component_factory = component_type2factory[component_type]
    generator_factory = component_type2factory[components_generator]
    components = generator_factory(component_factory, doe)
    return components


if __name__ == "__main__":
    pass



import pp
from pp.component import Component
from typing import Any, List, Tuple


@pp.autoname
def mmi1x2(
    wg_width: float = 0.5,
    width_taper: float = 1.0,
    length_taper: float = 10.0,
    length_mmi: float = 5.496,
    width_mmi: float = 2.5,
    gap_mmi: float = 0.25,
    layer: Tuple[int, int] = pp.LAYER.WG,
    layers_cladding: List[Any] = [],
    cladding_offset: float = 3.0,
) -> Component:
    """mmi 1x2

    Args:
        wg_width: input waveguides width
        width_taper: interface between input waveguides and mmi region
        length_taper: into the mmi region
        length_mmi: in x direction
        width_mmi: in y direction
        gap_mmi:  gap between tapered wg
        layer: gds layer

    .. plot::
      :include-source:

      import pp
      c = pp.c.mmi1x2(width_mmi=2, length_mmi=2.8)
      pp.plotgds(c)

    """
    c = pp.Component()
    c.wg_width = wg_width
    w_mmi = width_mmi
    w_taper = width_taper

    taper = pp.c.taper(
        length=length_taper,
        width1=wg_width,
        width2=w_taper,
        layer=layer,
        layers_cladding=layers_cladding,
        cladding_offset=cladding_offset,
    )

    a = gap_mmi / 2 + width_taper / 2
    mmi = c << pp.c.rectangle(
        size=(length_mmi, w_mmi),
        layer=layer,
        ports_parameters={
            "E": [(w_mmi / 2 - a, w_taper), (w_mmi / 2 + a, w_taper)],
            "W": [(w_mmi / 2, w_taper)],
        },
    )
    mmi.y = 0

    for layer_cladding in layers_cladding:
        clad = c << pp.c.rectangle(
            size=(length_mmi, w_mmi + 2 * cladding_offset), layer=layer_cladding
        )
        clad.y = 0
        c.absorb(clad)

    # For each port on the MMI rectangle
    for port_name, port in mmi.ports.items():

        # Create a taper
        taper_ref = c.add_ref(taper)

        # Connect the taper to the mmi section
        taper_ref.connect(port="2", destination=mmi.ports[port_name])

        # Add the taper port
        c.add_port(name=port_name, port=taper_ref.ports["1"])
        c.absorb(taper_ref)

    c.move(origin=c.ports["W0"].position, destination=(0, 0))
    c.simulation_settings = dict(port_width=1.5e-6)
    c.absorb(mmi)

    return c


@pp.autoname
def mmi1x2_biased(
    wg_width=0.5,
    width_taper=1.0,
    length_taper=10,
    length_mmi=5.496,
    width_mmi=2.5,
    gap_mmi=0.25,
    layer=pp.LAYER.WG,
):
    return mmi1x2(
        wg_width=pp.bias.width(wg_width),
        width_taper=pp.bias.width(width_taper),
        length_taper=length_taper,
        length_mmi=length_mmi,
        width_mmi=pp.bias.width(width_mmi),
        gap_mmi=pp.bias.gap(gap_mmi),
        layer=layer,
    )


if __name__ == "__main__":
    c = mmi1x2(pins=True)
    print(c.ports)
    # print(c.get_ports_array())
    # c = mmi1x2_biased()
    # pp.write_to_libary("mmi1x2", width_mmi=10, overwrite=True)
    # print(c.get_optical_ports())
    pp.write_gds(c, pp.CONFIG["gdsdir"] / "mmi1x2.gds")
    pp.show(c)
    # print(c.get_settings())



"""
you can define both DOEs and placer information in a YAML file
all the placer information need to be nested inside a placer section

iso_lines_coarse1:
  component: ISO_COARS_OPT1
  settings:
    dx: [50.]

  placer:
    type: pack_row/ pack_col/ grid / fixed_coords
        pack_grid:
    x0: 0
    y0: 0
    align_x: W
    align_y: S
    next_to: iso_lines_coarse1

"""


import os
import sys
import collections
import numpy as np
from omegaconf import OmegaConf
import klayout.db as pya

import pp.autoplacer.text as text
from pp.autoplacer.helpers import import_cell, load_gds, CELLS
from pp.config import CONFIG

UM_TO_GRID = 1e3
DEFAULT_BBOX_LAYER_IGNORE = [(8484, 8484)]


def _print(*args, **kwargs):
    print(*args, **kwargs)
    sys.stdout.flush()


def to_grid(x, um_to_grid=UM_TO_GRID):
    return int(x * um_to_grid)


class SizeInfo:
    def __init__(
        self,
        cell,
        layout=None,
        ignore_layers=DEFAULT_BBOX_LAYER_IGNORE,
        um_to_grid=UM_TO_GRID,
    ):
        """
        layout is required if cell is a cell reference instead of a cell
        """
        if isinstance(cell, pya.CellInstArray):
            bbox = cell.bbox(layout)
        else:
            parent_layout = cell.layout()
            layer_indexes = parent_layout.layer_indexes()

            for process_id, purpose_id in ignore_layers:
                layer_id = parent_layout.find_layer(process_id, purpose_id)
                if layer_id in layer_indexes:
                    layer_indexes.remove(layer_id)

            bbox = None
            for layer_id in layer_indexes:
                per_layer_bbox = cell.bbox_per_layer(layer_id)
                if bbox is None:
                    bbox = per_layer_bbox
                else:
                    bbox = bbox + per_layer_bbox

        self.box = bbox

        self.west = self.box.left / um_to_grid
        self.east = self.box.right / um_to_grid

        self.south = self.box.bottom / um_to_grid
        self.north = self.box.top / um_to_grid

        self.width = self.east - self.west
        self.height = self.north - self.south

        xc = int(0.5 * (self.east + self.west))
        yc = int(0.5 * (self.north + self.south))

        self.sw = np.array([self.west, self.south])
        self.se = np.array([self.east, self.south])
        self.nw = np.array([self.west, self.north])
        self.ne = np.array([self.east, self.north])

        self.cw = np.array([self.west, yc])
        self.ce = np.array([self.east, yc])
        self.nc = np.array([xc, self.north])
        self.sc = np.array([xc, self.south])
        self.cc = self.center = np.array([xc, yc])

    @property
    def rect(self):
        w, e, s, n = self.west, self.east, self.south, self.north
        return [(w, s), (e, s), (e, n), (w, n)]

    def __str__(self):
        return "w: {}\ne: {}\ns: {}\nn: {}\n".format(
            self.west, self.east, self.south, self.north
        )


def placer_grid_cell_refs(
    cells,
    cols=1,
    rows=1,
    dx=10.0,
    dy=10.0,
    x0=0,
    y0=0,
    um_to_grid=UM_TO_GRID,
    **settings,
):
    """cells: list of cells - order matters for placing"""

    indices = [(i, j) for j in range(cols) for i in range(rows)]

    if rows * cols < len(cells):
        raise ValueError(
            "Shape ({}, {}): Not enough emplacements ({}) for all these components"
            " ({}).".format(rows, cols, len(indices), len(cells))
        )
    components = []
    for cell, (i, j) in zip(cells, indices):
        _x = int((x0 + j * dx) * um_to_grid)
        _y = int((y0 + i * dy) * um_to_grid)

        transform = pya.Trans(_x, _y)
        c_ref = pya.CellInstArray(cell.cell_index(), transform)
        components += [c_ref]

    return components


def pack_row(
    cells,
    row_ids=None,
    nb_cols=None,
    x0=0,
    y0=0,
    align_x="W",
    align_y="S",
    margin=20,
    margin_x=None,
    margin_y=None,
    um_to_grid=UM_TO_GRID,
    period_x=None,
    period_y=None,
    rotation=0,
):
    """
    Args:
        cells: a list of cells  (size n)
        row_ids: a list of row ids (size n)
            where each id represents the row where the cell should be placed
            None by default => all cells in the same row


        period_x, period_y: not used by default,
            if set, use this period instead of computing the component spacing
            from the margin and the component dimension


    returns a list of cell references
    """
    si_list = [SizeInfo(c, um_to_grid=um_to_grid) for c in cells]
    heights = [si.height for si in si_list]
    margin_y = margin_y if margin_y is not None else margin
    margin_x = margin_x if margin_x is not None else margin

    if row_ids is None:
        row_ids = []
        nb_cells = len(cells)
        if nb_cols is None:
            nb_cols = len(cells)
        nb_full_rows = nb_cells // nb_cols
        nb_cols_last_row = nb_cells % nb_cols
        for row_id in range(nb_full_rows):
            row_ids += [row_id] * nb_cols

        last_row_index = row_id + 1
        row_ids += [last_row_index] * nb_cols_last_row

    if len(cells) != len(row_ids):
        raise ValueError(
            "Each cell should be assigned a row id.         Got {} cells for {} row ids".format(
                len(cells), len(row_ids)
            )
        )

    # Find the height of each row to fit the cells
    # Also group the cells by row

    unique_row_ids = list(set(row_ids))
    unique_row_ids.sort()
    _row_to_heights = {r: [] for r in set(row_ids)}
    row_to_cells = {r: [] for r in unique_row_ids}
    for row, h, cell in zip(row_ids, heights, cells):
        _row_to_heights[row] += [h]
        row_to_cells[row] += [cell]

    row_to_height = {k: max(v) for k, v in _row_to_heights.items()}

    components = []

    # Do the packing per row
    y = y0
    for row in unique_row_ids:
        cells = row_to_cells[row]
        x = x0

        for c in cells:
            si = SizeInfo(c, um_to_grid=um_to_grid)
            if align_x == "W" and align_y == "S":
                component_origin = si.sw
            elif align_x == "E" and align_y == "S":
                component_origin = si.se
            elif align_x == "E" and align_y == "N":
                component_origin = si.ne
            elif align_x == "W" and align_y == "N":
                component_origin = si.nw
            try:
                _x = to_grid(x - component_origin[0], um_to_grid)
                _y = to_grid(y - component_origin[1], um_to_grid)

                transform = pya.Trans(rotation / 2, 0, _x, _y)
                # transform = pya.Trans(_x, _y)
                c_ref = pya.CellInstArray(c.cell_index(), transform)
                components += [c_ref]

            except:
                print(x, component_origin[0], um_to_grid)
                print("ISSUE PLACING AT", _x, _y)
                if align_x not in ["W", "E"]:
                    _print("align_x should be `W`, `E` or a float")
                if align_y not in ["N", "S"]:
                    _print("align_y should be `N`, `S` or a float")
                # raise

            dx = si.width + margin_x if period_x is None else period_x
            if align_x == "W":
                x += dx
            else:
                x -= dx

        dy = row_to_height[row] + margin_y if period_y is None else period_y

        if align_y == "S":
            y += dy
        else:
            y -= dy

    return components


def pack_col(
    cells,
    col_ids=None,
    nb_rows=None,
    x0=0,
    y0=0,
    align_x="W",
    align_y="S",
    margin=20,
    margin_x=None,
    margin_y=None,
    um_to_grid=UM_TO_GRID,
    period_x=None,
    period_y=None,
):
    """
    Args:
        cells: a list of cells  (size n)
        col_ids: a list of column ids (size n)
            where each id represents the row where the cell should be placed
            None by default => all cells are packed in the same column

    returns a list of cell references
    """
    widths = [SizeInfo(c, um_to_grid=um_to_grid).width for c in cells]
    margin_y = margin_y if margin_y is not None else margin
    margin_x = margin_x if margin_x is not None else margin

    if col_ids is None:
        col_ids = []
        nb_cells = len(cells)
        if nb_rows is None:
            nb_rows = len(cells)
        nb_full_cols = nb_cells // nb_rows
        nb_rows_last_col = nb_cells % nb_rows
        for col_id in range(nb_full_cols):
            col_ids += [col_id] * nb_rows

        last_col_index = col_id + 1
        col_ids += [last_col_index] * nb_rows_last_col

    if len(cells) != len(col_ids):
        raise ValueError(
            "Each cell should be assigned a row id.         Got {} cells for {} col ids".format(
                len(cells), len(col_ids)
            )
        )

    # Find the width of each column to fit the cells
    # Also group the cells by column

    unique_col_ids = list(set(col_ids))
    unique_col_ids.sort()
    _col_to_widths = {r: [] for r in set(col_ids)}
    col_to_cells = {r: [] for r in unique_col_ids}
    for col, w, cell in zip(col_ids, widths, cells):
        _col_to_widths[col] += [w]
        col_to_cells[col] += [cell]

    col_to_width = {k: max(v) for k, v in _col_to_widths.items()}

    components = []

    # Do the packing per column
    x = x0
    for col in unique_col_ids:
        cells = col_to_cells[col]
        y = y0

        for c in cells:
            si = SizeInfo(c, um_to_grid=um_to_grid)
            if align_x == "W" and align_y == "S":
                component_origin = si.sw
            elif align_x == "E" and align_y == "S":
                component_origin = si.se
            elif align_x == "E" and align_y == "N":
                component_origin = si.ne
            elif align_x == "W" and align_y == "N":
                component_origin = si.nw

            _x = to_grid(x - component_origin[0], um_to_grid=um_to_grid)
            _y = to_grid(y - component_origin[1], um_to_grid=um_to_grid)

            try:
                transform = pya.Trans(_x, _y)
                c_ref = pya.CellInstArray(c.cell_index(), transform)
                components += [c_ref]
            except:
                print(x, component_origin[0], um_to_grid)
                print(y, component_origin[1], um_to_grid)
                print("ISSUE PLACING AT", _x, _y)
                print("ISSUE PLACING at", _x, _y)

            dy = si.height + margin_y if period_y is None else period_y
            if align_y == "S":
                y += dy
            else:
                y -= dy

        dx = col_to_width[col] + margin_x if period_x is None else period_x
        if align_x == "W":
            x += dx
        else:
            x -= dx

    return components


def placer_fixed_coords(
    cells, x, y, x0=0, y0=0, do_permutation=False, um_to_grid=UM_TO_GRID, **kwargs
):
    """place cells using a list of coordinates"""

    # List all coordinates
    if do_permutation:
        coords = [(_x, _y) for _x in x for _y in y]
    else:
        coords = [(_x, _y) for _x, _y in zip(x, y)]

    # Update origin
    coords = [(c[0] + x0, c[1] + y0) for c in coords]

    # Generate cell list
    if len(cells) == 1:
        cells = cells * len(coords)

    # update coordinates from um to grid
    coords = [(int(c[0] * um_to_grid), int(c[1] * um_to_grid)) for c in coords]

    # Generate transforms
    transforms = [pya.Trans(*c) for c in coords]

    return [pya.CellInstArray(c.cell_index(), t) for c, t in zip(cells, transforms)]


def load_yaml(filepath, defaults={"do_permutation": True}):
    """load placer settings

    Args:
        filepath: a yaml file containing the does and placer information

    Returns:
        a dictionnary of DOEs with:
        {
            doe_name1: {...}
            doe_name2: {...}
            ...
        }

    """

    does = {}
    data = OmegaConf.load(filepath)
    data = OmegaConf.to_container(data)
    mask = data.pop("mask")

    if "layer_doe_label" not in mask:
        mask["layer_doe_label"] = (102, 6)

    for doe_name, doe in data.items():
        # do_permutation = defaults["do_permutation"]
        # if "do_permutation" in doe:
        # do_permutation = doe.pop("do_permutation")
        _doe = {}
        _doe.update(doe)
        does[doe_name] = _doe
    return does, mask


DOE_CELLS = {}


def load_doe(doe_name, doe_root):
    """
    Load all components for this DOE from the cache
    """
    doe_dir = os.path.join(doe_root, doe_name)
    content_file = os.path.join(doe_dir, "content.txt")

    if os.path.isfile(content_file):
        with open(content_file) as f:
            lines = f.read().split("\n")
            line = lines[0]

            if line.startswith("TEMPLATE:"):
                """
                If using a template, load the GDS from DOE folder used as a template
                """
                template_name = line.split(":")[1].strip()
                return load_doe(template_name, doe_root)

            else:
                """
                Otherwise load the GDS from the current folder
                """
                component_names = line.split(" , ")
                gdspaths = [
                    os.path.join(doe_dir, name + ".gds") for name in component_names
                ]
                cells = [load_gds(gdspath) for gdspath in gdspaths]

        # print("LOAD DOE")
        # for _c in cells:
        # print(_c.top_cell().name)
        # print()

        return cells


PLACER_NAME2FUNC = {
    "grid": placer_grid_cell_refs,
    "pack_row": pack_row,
    "pack_col": pack_col,
    "fixed_coords": placer_fixed_coords,
}


def separate_does_from_templates(dicts):
    templates = {}
    does = {}
    for name, d in dicts.items():
        if "type" in d.keys():  # and d["type"] == "template":
            templates[name] = d
        else:
            does[name] = d

    # We do not want to propagate "type": template to the does => removing it here
    for d in templates.values():
        d.pop("type")

    return does, templates


def update_dicts_recurse(target_dict, default_dict):
    target_dict = target_dict.copy()
    default_dict = default_dict.copy()
    for k, v in default_dict.items():
        if k not in target_dict:
            vtype = type(v)
            if vtype == dict or vtype == collections.OrderedDict:
                target_dict[k] = v.copy()  # To avoid issues when popping
            else:
                target_dict[k] = v
        else:
            vtype = type(target_dict[k])
            if vtype == dict or vtype == collections.OrderedDict:
                target_dict[k] = update_dicts_recurse(target_dict[k], default_dict[k])
    return target_dict


def place_from_yaml(
    filepath_yaml,
    root_does=CONFIG["cache_doe_directory"],
    precision=1e-9,
    fontpath=text.FONT_PATH,
):
    """Returns a gds cell composed of DOEs/components given in a yaml file
    allows for each DOE to have its own x and y spacing (more flexible than method1)

    Args:
        filepath_yaml:
        root_does: used for cache, requires content.txt
    """
    transform_identity = pya.Trans(0, 0)
    dicts, mask_settings = load_yaml(filepath_yaml)

    does, templates = separate_does_from_templates(dicts)

    placed_doe = None
    placed_does = {}
    top_level_name = mask_settings.get("name", "TOP_LEVEL")
    layer_doe_label = mask_settings["layer_doe_label"]
    top_level_layout = pya.Layout()

    # Set database units according to precision
    top_level_layout.dbu = precision / 1e-6
    dbu = top_level_layout.dbu
    um_to_grid = int(1 / dbu)

    top_level = top_level_layout.create_cell(top_level_name)
    global CELLS
    CELLS[top_level_name] = top_level_layout

    default_doe_settings = {
        "add_doe_label": False,
        "add_doe_visual_label": False,
        "dx_visual_label": 0,
        "dy_visual_label": 0,
    }

    for doe_name, doe in does.items():

        # If a template is specified, apply it
        if "template" in doe:
            doe_templates = doe["template"]
            if type(doe_templates) != list:
                doe_templates = [doe_templates]
            for doe_template in doe_templates:
                try:
                    doe = update_dicts_recurse(doe, templates[doe_template])

                except:
                    print(doe_template, "does not exist")
                    raise
        doe = update_dicts_recurse(doe, default_doe_settings)

        # Get all the components
        components = load_doe(doe_name, root_does)

        """
        # Check that the high level components are all unique
        # For now this is mostly to circumvent a bug
        # But the design manual also specifies that DOE components should have
        # unique names. So one instance per cell
        """

        if components:
            if len(components) != len(set([_c.top_cell().name for _c in components])):
                __dict_component_debug = {}
                for _c in components:
                    _name = _c.top_cell().name
                    if _name not in __dict_component_debug:
                        __dict_component_debug[_name] = 0
                    __dict_component_debug[_name] += 1
                duplicates_components = [
                    _name
                    for _name, _count in __dict_component_debug.items()
                    if _count > 1
                ]
                print("Please remove duplicate components at DOE entry level: ")
                print(duplicates_components)

            components = [
                import_cell(top_level_layout, _c.top_cell()) for _c in components
            ]

        # Find placer information
        default_placer_settings = {
            "align_x": "W",
            "align_y": "S",
            "margin": 10,
            "x0": "E",
            "y0": "S",
        }
        settings = default_placer_settings.copy()
        placer = doe.get("placer")

        if placer:
            placer_type = placer.pop("type", "pack_col")
            settings.update(doe["placer"])
        else:
            placer_type = "pack_col"

        if placer_type not in PLACER_NAME2FUNC:
            raise ValueError(
                f"{placer_type} is not an available placer, Choose:"
                f" {list(PLACER_NAME2FUNC.keys())}"
            )
        _placer = PLACER_NAME2FUNC[placer_type]

        # All other attributes are assumed to be settings for the placer

        ## Check if the cell should be attached to a specific parent cell
        if "parent" in settings:
            parent_name = settings.pop("parent")
            if parent_name not in CELLS:
                # Create parent cell in layout and insert it under top level
                parent_cell = top_level_layout.create_cell(parent_name)
                CELLS[parent_name] = parent_cell
                parent_cell_instance = pya.CellInstArray(
                    parent_cell.cell_index(), transform_identity
                )
                top_level.insert(parent_cell_instance)
            doe_parent_cell = CELLS[parent_name]
        else:
            # If no parent specified, insert the DOE at top level
            doe_parent_cell = top_level

        ## Check if we should create a DOE cell which regroups the DOEs
        if "with_doe_cell" in settings:
            with_doe_cell = settings.pop("with_doe_cell")
        else:
            with_doe_cell = True

        # x0, y0 can either be float or string
        x0 = settings.pop("x0")
        y0 = settings.pop("y0")

        # Check whether we are doing relative or absolute placement
        # if (x0 in ["E", "W"] or y0 in ["N", "S"]) and not placed_doe:
        #     raise ValueError(
        #         "At least one DOE must be placed to use relative placement"
        #     )

        # For relative placement (to previous DOE)
        if "margin_x" not in settings:
            settings["margin_x"] = settings["margin"]
        if "margin_y" not in settings:
            settings["margin_y"] = settings["margin"]

        if "inter_margin_x" not in settings:
            inter_margin_x = settings["margin_x"]
        else:
            inter_margin_x = settings.pop("inter_margin_x")

        if "inter_margin_y" not in settings:
            inter_margin_y = settings["margin_y"]
        else:
            inter_margin_y = settings.pop("inter_margin_y")

        align_x = settings["align_x"]
        align_y = settings["align_y"]

        ## Making sure that the alignment is sensible depending on how we stack

        # If we specify a DOE to place next to, use it
        if "next_to" in settings:
            placed_doe = placed_does[settings.pop("next_to")]

        print(placed_doe)
        print(placed_does)

        # Otherwise, use previously placed DOE as starting point
        doe_si = (
            SizeInfo(placed_doe, top_level_layout, um_to_grid=um_to_grid)
            if placed_doe is not None
            else None
        )
        if x0 == "E":
            x0 = doe_si.east
            if align_x == "W":
                x0 += inter_margin_x

        if x0 == "W":
            x0 = doe_si.west
            if align_x == "E":
                x0 -= inter_margin_x

        if y0 == "N":
            y0 = doe_si.north
            if align_y == "S":
                y0 += inter_margin_y

        if y0 == "S":
            y0 = doe_si.south
            if align_y == "N":
                y0 -= inter_margin_y

        # Add x0, y0 in settings as float
        settings["x0"] = x0
        settings["y0"] = y0

        settings["um_to_grid"] = um_to_grid

        placed_components = _placer(components, **settings)

        # Place components within a cell having the DOE name

        if with_doe_cell or len(placed_components) > 1:
            doe_cell = top_level_layout.create_cell(doe_name)
            CELLS[doe_name] = doe_cell
            for instance in placed_components:
                doe_cell.insert(instance)
            placed_does[doe_name] = doe_cell
            placed_doe = doe_cell
            doe_instance = pya.CellInstArray(doe_cell.cell_index(), transform_identity)
        else:
            # If only single cell and we want to skip the doe cell
            doe_instance = placed_components[0]
            placed_does[doe_name] = doe_instance
            placed_doe = doe_instance

        add_doe_label = doe["add_doe_label"]
        add_doe_visual_label = doe["add_doe_visual_label"]

        if add_doe_label:
            label_layer_index, label_layer_datatype = layer_doe_label
            layer_index = top_level.layout().insert_layer(
                pya.LayerInfo(label_layer_index, label_layer_datatype)
            )
            # Add the name of the DOE at the center of the cell
            _p = doe_instance.bbox(top_level_layout).center()
            _text = pya.Text(doe_name, _p.x, _p.y)
            top_level.shapes(layer_index).insert(_text)

        if add_doe_visual_label:
            _bbox = doe_instance.bbox(top_level_layout)

            idbu = 1 / top_level.layout().dbu
            x_text = _bbox.center().x + doe["dx_visual_label"] * idbu
            y_text = _bbox.bottom + (15.0 + doe["dy_visual_label"]) * idbu
            _text = text.add_text(
                top_level, doe_name, position=(x_text, y_text), fontpath=fontpath
            )
            # _transform = pya.DTrans(x_text, y_text)
            # top_level.insert(pya.CellInstArray(_text.cell_index(), _transform))

        doe_parent_cell.insert(doe_instance)

    return top_level


def place_and_write(
    filepath_yaml, root_does=CONFIG["cache_doe_directory"], filepath_gds="top_level.gds"
):
    c = place_from_yaml(filepath_yaml, root_does)
    _print("writing...")
    c.write(filepath_gds)


def assemble_subdies_from_yaml(filepath, subdies_directory, mask_directory=None):
    data = OmegaConf.load(filepath)
    data = OmegaConf.to_container(data)

    mask = data.pop("mask")
    mask_name = mask["name"]

    # Remaining entries are subdies
    dict_subdies = {
        k: (v["x"], v["y"], v["R"] if "R" in v else 0) for k, v in data.items()
    }

    return assemble_subdies(mask_name, dict_subdies, subdies_directory, mask_directory)


def assemble_subdies(
    mask_name,
    dict_subdies,
    subdies_directory,
    mask_directory=None,
    um_to_grid=UM_TO_GRID,
):
    """
    Args:
        dict_subdies: {subdie_name: (x, y, rotation) in (um, um, deg)}
        subdies_directory: directory where the subdies should be looked for
    """
    top_level_layout = pya.Layout()
    top_level = top_level_layout.create_cell(mask_name)
    if mask_directory is None:
        mask_directory = subdies_directory

    for subdie_name, (x_um, y_um, R) in dict_subdies.items():
        gdspath = os.path.join(subdies_directory, subdie_name + ".gds")
        subdie = load_gds(gdspath).top_cell()

        _subdie = import_cell(top_level_layout, subdie)

        t = pya.Trans(R / 2, 0, int(x_um * um_to_grid), int(y_um * um_to_grid))
        # t = pya.Trans(0, 0)
        subdie_instance = pya.CellInstArray(_subdie.cell_index(), t)
        top_level.insert(subdie_instance)

    top_level.write(os.path.join(mask_directory, mask_name + ".gds"))
    return top_level


def _demo():
    import pp

    c = pp.c.waveguide()
    gdspath = pp.write_component(c)

    layout1 = load_gds(gdspath)
    cell1 = layout1.top_cell()
    cell1_instance1 = pya.CellInstArray(cell1.cell_index(), pya.Trans(10, 0))

    layout2 = pya.Layout()
    layout2.create_cell("TOP_LEVEL")

    layout2.cell("TOP_LEVEL").insert(cell1_instance1)
    layout2.write("test.gds")


if __name__ == "__main__":
    _demo()
    print(CELLS)



from pp.container import container
from pp.component import Component
from pp.routing.route_pad_array import route_pad_array
from pp.rotate import rotate


@container
def add_electrical_pads(component: Component, rotation=180, **kwargs):
    """add compnent with top electrical pads and routes
    Args:
        component: Component,
        pad_spacing: float = 150.,
        pad: Callable = pad,
        fanout_length: Optional[int] = None,
        max_y0_optical: None = None,
        waveguide_separation: float = 4.0,
        bend_radius: float = 0.1,
        connected_port_list_ids: None = None,
        n_ports: int = 1,
        excluded_ports: List[Any] = [],
        pad_indices: None = None,
        route_filter: Callable = connect_elec_waypoints,
        port_name: str = "W",
        pad_rotation: int = -90,
        x_pad_offset: int = 0,
        port_labels: None = None,
        select_ports: Callable = select_electrical_ports,

    """

    c = Component(f"{component.name}_pad")

    cr = component.rotate(rotation)

    elements, pads, _ = route_pad_array(component=cr, **kwargs,)

    c << cr
    for e in elements:
        c.add(e)
    for e in pads:
        c.add(e)

    for pname, p in cr.ports.items():
        if p.port_type == "optical":
            c.add_port(pname, port=p)

    return rotate(c, -rotation)


if __name__ == "__main__":
    import pp

    c = pp.c.cross(length=100, layer=pp.LAYER.M3, port_type="dc")
    c = pp.c.mzi2x2(with_elec_connections=True)
    c = pp.c.waveguide_heater()
    c.move((20, 50))
    cc = add_electrical_pads(c, fanout_length=100)
    print(cc.ports)

    ccc = pp.routing.add_fiber_array(cc)
    pp.show(ccc)



import pp
from pp.components.bend_s import bend_s
from pp.components.waveguide import waveguide
from pp.component import Component
from typing import Callable


@pp.autoname
def coupler_asymmetric(
    bend: Callable = bend_s,
    waveguide: Callable = waveguide,
    gap: float = 0.234,
    wg_width: float = 0.5,
) -> Component:
    """ bend coupled to straight waveguide

    Args:
        bend:
        waveguide: waveguide factory
        gap: um
        wg_width

    .. plot::
      :include-source:

      import pp

      c = pp.c.coupler_asymmetric()
      pp.plotgds(c)

    """
    bend = pp.call_if_func(bend, width=wg_width)
    wg = pp.call_if_func(waveguide, width=wg_width)

    w = bend.ports["W0"].width
    y = (w + gap) / 2

    c = pp.Component()
    wg = wg.ref(position=(0, y), port_id="W0")
    bottom_bend = bend.ref(position=(0, -y), port_id="W0", v_mirror=True)

    c.add(wg)
    c.add(bottom_bend)

    # Using absorb here to have a flat cell and avoid
    # to have deeper hierarchy than needed
    c.absorb(wg)
    c.absorb(bottom_bend)

    port_width = 2 * w + gap
    c.add_port(name="W0", midpoint=[0, 0], width=port_width, orientation=180)
    c.add_port(port=bottom_bend.ports["E0"], name="E0")
    c.add_port(port=wg.ports["E0"], name="E1")

    return c


if __name__ == "__main__":
    c = coupler_asymmetric(wg_width=0.6, gap=0.4)
    pp.show(c)



import pp
from pp.components.coupler90 import coupler90
from pp.components.waveguide import waveguide
from pp.components.coupler_straight import coupler_straight
from pp.netlist_to_gds import netlist_to_component
from pp.name import autoname
from pp.drc import assert_on_2nm_grid
from pp.component import Component
from typing import Callable, Dict, List, Tuple


@autoname
def ring_double_bus(**kwargs) -> Component:
    """ Ring double bus

    Args:
        bend_radius=5
        length_x=1
        length_y=1
        gap=0.2
        coupler90_factory=coupler90
        straight_factory=waveguide
        cpl_straight_factory=coupler_straight
        wg_width=0.5


    .. code::

        -CTL==CTS==CTR-
         |          |
         VL         VR
         |          |
        -CBL==CBS==CBR-

    .. plot::
      :include-source:

      import pp

      c = pp.c.ring_double_bus(gap=0.2, length_x=10, length_y=5, bend_radius=5)
      pp.plotgds(c)


    """
    components, connections, ports_map = ring_double_bus_netlist(**kwargs)
    component = netlist_to_component(components, connections, ports_map)
    return component


def ring_double_bus_netlist(
    bend_radius: float = 5.0,
    length_x: float = 1.0,
    length_y: float = 1.0,
    gap: float = 0.2,
    coupler90_factory: Callable = coupler90,
    straight_factory: Callable = waveguide,
    cpl_straight_factory: Callable = coupler_straight,
    wg_width: float = 0.5,
) -> Tuple[
    Dict[str, Tuple[Component, str]],
    List[Tuple[str, str, str, str]],
    Dict[str, Tuple[str, str]],
]:
    """
    .. code::

         -CTL==CTS==CTR-
         |          |
         VL         VR
         |          |
        -CBL==CBS==CBR-
    """
    assert_on_2nm_grid(gap)

    _cpl_bend = coupler90_factory(bend_radius=bend_radius, width=wg_width, gap=gap)

    _cpl_straight = cpl_straight_factory(length=length_x, gap=gap, width=wg_width)
    _v = straight_factory(length=length_y)

    components = {
        "CBL": (_cpl_bend, "mirror_y"),
        "CBR": (_cpl_bend, "None"),
        "CTR": (_cpl_bend, "mirror_x"),
        "CTL": (_cpl_bend, "R180"),
        "CTS": (_cpl_straight, "None"),
        "CBS": (_cpl_straight, "None"),
        "VL": (_v, "R90"),
        "VR": (_v, "R90"),
    }

    connections = [
        ("CBL", "W0", "CBS", "W0"),
        ("CBS", "E0", "CBR", "W0"),
        ("CBL", "N0", "VL", "W0"),
        ("VL", "E0", "CTL", "N0"),
        ("CTL", "W0", "CTS", "W0"),
        ("CTS", "E0", "CTR", "W0"),
        ("CBR", "N0", "VR", "W0"),
        ("VR", "E0", "CTR", "N0"),
    ]

    ports_map = {
        "W0": ("CBL", "E0"),
        "E0": ("CBR", "E0"),
        "W1": ("CTL", "E0"),
        "E1": ("CTR", "E0"),
    }

    return components, connections, ports_map


if __name__ == "__main__":
    c = ring_double_bus(bend_radius=5.0, length_x=2, length_y=4, gap=0.2)
    pp.show(c)



from typing import Callable
import numpy as np
import scipy.optimize as so
import gdspy

import pp
from pp.components import taper
from pp.components.bezier import bezier
from pp.components.bezier import bezier_curve
from pp.components.bezier import find_min_curv_bezier_control_points
from pp.geo_utils import path_length
from pp.config import GRID_PER_UNIT
from pp.components.ellipse import ellipse
from pp.layers import LAYER
from pp.component import Component


def rnd(p):
    """ round """
    return np.round(p * GRID_PER_UNIT) / GRID_PER_UNIT


@pp.autoname
def crossing_arm(
    wg_width: float = 0.5,
    r1: float = 3.0,
    r2: float = 1.1,
    w: float = 1.2,
    L: float = 3.4,
) -> Component:
    """ arm of a crossing
    """
    c = pp.Component()
    _ellipse = ellipse(radii=(r1, r2), layer=LAYER.SLAB150).ref()
    c.add(_ellipse)
    c.absorb(_ellipse)

    a = pp.drc.snap_to_1nm_grid(L + w / 2)
    h = wg_width / 2

    taper_pts = [
        (-a, h),
        (-w / 2, w / 2),
        (w / 2, w / 2),
        (a, h),
        (a, -h),
        (w / 2, -w / 2),
        (-w / 2, -w / 2),
        (-a, -h),
    ]

    c.add_polygon(taper_pts, layer=LAYER.WG)
    c.add_port(
        name="W0", midpoint=(-a, 0), orientation=180, width=wg_width, layer=LAYER.WG
    )

    c.add_port(
        name="E0", midpoint=(a, 0), orientation=0, width=wg_width, layer=LAYER.WG
    )

    return c


@pp.autoname
def crossing(arm: Callable = crossing_arm) -> Component:
    """waveguide crossing

    .. plot::
      :include-source:

      import pp

      c = pp.c.crossing()
      pp.plotgds(c)

    """
    cx = pp.Component()
    arm = pp.call_if_func(arm)
    arm_h = arm.ref()
    arm_v = arm.ref(rotation=90)

    port_id = 0
    for c in [arm_h, arm_v]:
        cx.add(c)
        cx.absorb(c)
        for p in c.ports.values():
            cx.add_port(name="{}".format(port_id), port=p)
            port_id += 1
    cx = pp.port.rename_ports_by_orientation(cx)
    return cx


@pp.autoname
def crossing_from_taper(taper=lambda: taper(width2=2.5, length=3.0)):
    """
    Crossing based on a taper. The default is a dummy taper
    """
    taper = pp.call_if_func(taper)

    c = pp.Component()
    for i, a in enumerate([0, 90, 180, 270]):
        _taper = taper.ref(position=(0, 0), port_id="2", rotation=a)
        c.add(_taper)
        c.add_port(name="{}".format(i), port=_taper.ports["1"])
        c.absorb(_taper)

    c = pp.port.rename_ports_by_orientation(c)
    return c


@pp.autoname
def crossing_etched(
    wg_width=0.5,
    r1=3.0,
    r2=1.1,
    w=1.2,
    L=3.4,
    layer_wg=LAYER.WG,
    layer_slab=LAYER.SLAB150,
):
    """
    Waveguide crossing:
    - The full crossing has to be on WG layer (to start with a 220nm slab)
    - Then we etch the ellipses down to 150nm slabs and we keep linear taper at 220nm. What we write is what we etch on this step
    """

    # Draw the ellipses
    c = pp.Component()
    _ellipse1 = ellipse(radii=(r1, r2), layer=layer_wg).ref()
    _ellipse2 = ellipse(radii=(r2, r1), layer=layer_wg).ref()
    c.add(_ellipse1)
    c.add(_ellipse2)
    c.absorb(_ellipse1)
    c.absorb(_ellipse2)

    #
    a = L + w / 2
    h = wg_width / 2

    taper_cross_pts = [
        (-a, h),
        (-w / 2, w / 2),
        (-h, a),
        (h, a),
        (w / 2, w / 2),
        (a, h),
        (a, -h),
        (w / 2, -w / 2),
        (h, -a),
        (-h, -a),
        (-w / 2, -w / 2),
        (-a, -h),
    ]

    tapers_poly = c.add_polygon(taper_cross_pts, layer=layer_wg)

    b = a - 0.1  # To make sure we get 4 distinct polygons when doing bool ops
    tmp_polygon = [(-b, b), (b, b), (b, -b), (-b, -b)]

    polys_etch = gdspy.fast_boolean([tmp_polygon], tapers_poly, "not", layer=layer_slab)
    c.add(polys_etch)

    positions = [(a, 0), (0, a), (-a, 0), (0, -a)]
    angles = [0, 90, 180, 270]

    i = 0
    for p, angle in zip(positions, angles):
        c.add_port(
            name="tmp{}".format(i),
            midpoint=p,
            orientation=angle,
            width=wg_width,
            layer=layer_wg,
        )
        i += 1

    c = pp.port.rename_ports_by_orientation(c)
    return c


@pp.autoname
def crossing45(crossing=crossing, port_spacing=40.0, dx=None, alpha=0.08):
    r""" 45Deg crossing with bends

    Args:
        crossing: 90D crossing
        port_spacing: target I/O port spacing
        dx: target length
        alpha: optimization parameter. Try with 0.1 to start with.
            - If the structure has too tight bends, diminish it.
            - If raise assertion angle errors, increase it


    Implementation note: The 45 Degree crossing CANNOT be kept as an SRef since
    we only allow for multiples of 90Deg rotations in SRef

    .. code::

        ----   ----
            \ /
             X
            / \
        ---    ----

    """

    # Instantiate the crossing if it is a factory
    crossing = pp.call_if_func(crossing)

    c = pp.Component()

    _crossing = crossing.ref(rotation=45)
    c.add(_crossing)

    # Add bends
    p_e = _crossing.ports["E0"].midpoint
    p_w = _crossing.ports["W0"].midpoint
    p_n = _crossing.ports["N0"].midpoint
    p_s = _crossing.ports["S0"].midpoint

    # Flatten the crossing - not an SRef anymore
    c.absorb(_crossing)
    if dx is None:
        dx = port_spacing
    dy = port_spacing / 2

    t = np.linspace(0, 1, 101)

    start_angle = 45
    end_angle = 0
    cpts = find_min_curv_bezier_control_points(
        start_point=p_e,
        end_point=(dx, dy),
        start_angle=start_angle,
        end_angle=end_angle,
        t=t,
        alpha=alpha,
    )

    _bez_bend = bezier(
        control_points=cpts, t=t, start_angle=start_angle, end_angle=end_angle
    )

    tol = 1e-2
    assert abs(_bez_bend.info["start_angle"] - start_angle) < tol, _bez_bend.info[
        "start_angle"
    ]
    assert abs(_bez_bend.info["end_angle"] - end_angle) < tol, _bez_bend.info[
        "end_angle"
    ]

    # print(abs(_bez_bend.info["start_angle"] - start_angle))
    # print(abs(_bez_bend.info["end_angle"] - end_angle))

    b_tr = _bez_bend.ref(position=p_e, port_id="0")
    b_tl = _bez_bend.ref(position=p_n, port_id="0", h_mirror=True)
    b_bl = _bez_bend.ref(position=p_w, port_id="0", rotation=180)
    b_br = _bez_bend.ref(position=p_s, port_id="0", v_mirror=True)

    for cmp_ref in [b_tr, b_br, b_tl, b_bl]:
        # cmp_ref = _cmp.ref()
        c.add(cmp_ref)
        c.absorb(cmp_ref)

    c.info["components"] = {"bezier_bend": _bez_bend, "crossing": crossing}

    c.info["min_bend_radius"] = b_br.info["min_bend_radius"]
    # print (b_tr.info["min_bend_radius"])
    c.add_port("E0", port=b_br.ports["1"])
    c.add_port("E1", port=b_tr.ports["1"])
    c.add_port("W0", port=b_bl.ports["1"])
    c.add_port("W1", port=b_tl.ports["1"])
    c.snap_ports_to_grid()
    return c


@pp.autoname
def compensation_path(crossing45=crossing45, direction="top"):
    r""" Path with same path length as crossing45 but with input and output ports having same y coordinates

    Args:
        name: component name
        crossing45: the crossing45 component that we want to match in path length
            This component needs to have .info["components"] with bends and crossing
        direction: the direction in which the bend should go "top" / "bottom"

    Returns:
        <pp.Component> a compensation path



    crossing45:

    .. code::

          ----       ----
              \     /
               \   /
                \ /
                 X
                / \
               /   \
              /     \
          ----       ----

    Compensation path:

    .. code::

             --+--
           _/     \_
        --/         \--


    """
    # Get total path length taken by the bends
    crossing45 = pp.call_if_func(crossing45)
    X45_cmps = crossing45.info["components"]
    length = 2 * X45_cmps["bezier_bend"].info["length"]

    # Find a bezier S-bend with half this length, but with a fixed length
    # governed by the crossing45 X-distance (west to east ports) and
    # the crossing x_distance

    target_bend_length = length / 2

    def get_x_span(cmp):
        return cmp.ports["E0"].x - cmp.ports["W0"].x

    x_span_crossing45 = get_x_span(crossing45)
    x_span_crossing = get_x_span(X45_cmps["crossing"])

    # x span allowed for the bend
    x0 = (x_span_crossing45 - x_span_crossing) / 2

    def get_control_pts(x, y):
        return [(0, 0), (x0 / 2, 0), (x0 / 2, y), (x0, y)]

    def f(y):
        control_points = get_control_pts(x0, y)
        t = np.linspace(0, 1, 51)
        path_points = bezier_curve(t, control_points)
        return path_length(path_points) - target_bend_length

    """
    # We know that the path length of the s-bend between two ports
    p0 and p1 is :
    # - larger than the euclidian distance L2(p0, p1)
    # - smaller than the manhattan distance DL(p0, p1)
    #
    # This gives the bounds for the brentq root finding
    """

    ya = target_bend_length - x0
    yb = np.sqrt(target_bend_length ** 2 - x0 ** 2)

    solution = so.root_scalar(f, bracket=[ya, yb], method="brentq")

    y_bend = solution.root
    y_bend = rnd(y_bend)

    if direction == "top":
        v_mirror = False
    else:
        v_mirror = True

    sbend = bezier(control_points=get_control_pts(x0, y_bend))

    component = pp.Component()
    crossing0 = X45_cmps["crossing"].ref()
    component.add(crossing0)

    sbend_left = sbend.ref(
        position=crossing0.ports["W0"], port_id="1", v_mirror=v_mirror
    )
    sbend_right = sbend.ref(
        position=crossing0.ports["E0"], port_id="1", h_mirror=True, v_mirror=v_mirror
    )

    component.add(sbend_left)
    component.add(sbend_right)

    # Add ports
    component.add_port("W0", port=sbend_left.ports["0"])
    component.add_port("E0", port=sbend_right.ports["0"])

    component.info["min_bend_radius"] = sbend.info["min_bend_radius"]
    component.info["components"] = {"sbend": sbend}

    return component


def demo():
    """ plot curvature of bends
    FIXME: add more documentation
    """
    from matplotlib import pyplot as plt

    c = crossing45(port_spacing=20.0, dx=15)
    c2 = compensation_path(crossing45=c)
    print(c.info["min_bend_radius"])
    print(c2.info["min_bend_radius"])

    component = pp.Component(name="top_lvl")
    component.add(c.ref(port_id="W0"))
    component.add(c2.ref(port_id="W0", position=(0, 10)))

    bend_info1 = c.info["components"]["bezier_bend"].info
    bend_info2 = c2.info["components"]["sbend"].info

    DL = bend_info1["length"]
    L2 = bend_info1["length"]
    plt.plot(bend_info1["t"][1:-1] * DL, abs(bend_info1["curvature"]))
    plt.plot(bend_info2["t"][1:-1] * L2, abs(bend_info2["curvature"]))
    plt.xlabel("bend length (um)")
    plt.ylabel("curvature (um^-1)")
    pp.show(component)
    plt.show()


if __name__ == "__main__":
    # c = crossing()
    c = crossing45(port_spacing=15, pins=True)
    print(c.ports["E1"].y - c.ports["E0"].y)
    pp.show(c)
    # print(c.get_ports_array())
    # demo()
    # c = crossing_etched()
    # c = crossing_from_taper()



from typing import List, Tuple
import hashlib
import numpy as np
from scipy.special import binom
from scipy.optimize import minimize

import pp
from pp.layers import LAYER
from pp.geo_utils import extrude_path
from pp.geo_utils import angles_deg
from pp.geo_utils import snap_angle
from pp.geo_utils import path_length
from pp.geo_utils import curvature
from numpy import ndarray
from pp.component import Component


def bezier_curve(t: ndarray, control_points: List[Tuple[float, int]]) -> ndarray:
    xs = 0.0
    ys = 0.0
    n = len(control_points) - 1
    for k in range(n + 1):
        ank = binom(n, k) * (1 - t) ** (n - k) * t ** k
        xs += ank * control_points[k][0]
        ys += ank * control_points[k][1]

    return np.column_stack([xs, ys])


def bezier_points(control_points, width, t=np.linspace(0, 1, 101)):
    """ t: 1D array of points varying between 0 and 1
    """
    points = bezier_curve(t, control_points)
    return extrude_path(points, width)


def bezier_biased(width=0.5, **kwargs):
    width = pp.bias.width(width)
    return bezier(width=width, **kwargs)


# Not using autoname on bezier due to control_points and t spacing
def bezier(
    name: None = None,
    width: float = 0.5,
    control_points: List[Tuple[float, float]] = [
        (0.0, 0.0),
        (5.0, 0.0),
        (5.0, 2.0),
        (10.0, 2.0),
    ],
    t: ndarray = np.linspace(0, 1, 201),
    layer: Tuple[int, int] = LAYER.WG,
    **extrude_path_params
) -> Component:
    """ bezier bend """

    def _fmt_f(x):
        return "{:.3f}".format(x).rstrip("0").rstrip(".")

    def _fmt_cp(cps):
        return "_".join(["({},{})".format(_fmt_f(p[0]), _fmt_f(p[1])) for p in cps])

    if name is None:
        points_hash = hashlib.md5(_fmt_cp(control_points).encode()).hexdigest()
        name = "bezier_w{}_{}_{}".format(int(width * 1e3), points_hash, layer)

    c = pp.Component(name=name)
    path_points = bezier_curve(t, control_points)
    polygon_points = extrude_path(path_points, width, **extrude_path_params)
    angles = angles_deg(path_points)

    c.info["start_angle"] = angles[0]
    c.info["end_angle"] = angles[-2]

    a0 = angles[0] + 180
    a1 = angles[-2]

    a0 = snap_angle(a0)
    a1 = snap_angle(a1)

    p0 = path_points[0]
    p1 = path_points[-1]
    c.add_polygon(polygon_points, layer=layer)
    c.add_port(name="0", midpoint=p0, width=width, orientation=a0, layer=layer)
    c.add_port(name="1", midpoint=p1, width=width, orientation=a1, layer=layer)

    c.info["length"] = path_length(path_points)
    curv = curvature(path_points, t)
    c.info["min_bend_radius"] = 1 / max(np.abs(curv))
    c.info["curvature"] = curv
    c.info["t"] = t

    return c


def find_min_curv_bezier_control_points(
    start_point,
    end_point,
    start_angle,
    end_angle,
    t=np.linspace(0, 1, 201),
    alpha=0.05,
    nb_pts=2,
):
    def array_1d_to_cpts(a):
        xs = a[::2]
        ys = a[1::2]
        return [(x, y) for x, y in zip(xs, ys)]

    def objective_func(p):
        """
        We want to minimize a combination of:
            - max curvature
            - negligible mismatch with start angle and end angle
        """

        ps = array_1d_to_cpts(p)
        control_points = [start_point] + ps + [end_point]
        path_points = bezier_curve(t, control_points)

        max_curv = max(np.abs(curvature(path_points, t)))

        angles = angles_deg(path_points)
        dstart_angle = abs(angles[0] - start_angle)
        dend_angle = abs(angles[-2] - end_angle)
        angle_mismatch = dstart_angle + dend_angle
        return angle_mismatch * alpha + max_curv

    x0, y0 = start_point[0], start_point[1]
    xn, yn = end_point[0], end_point[1]

    initial_guess = []
    for i in range(nb_pts):
        x = (i + 1) * (x0 + xn) / (nb_pts)
        y = (i + 1) * (y0 + yn) / (nb_pts)
        initial_guess += [x, y]

    # initial_guess = [(x0 + xn) / 2, y0, (x0 + xn) / 2, yn]

    res = minimize(objective_func, initial_guess, method="Nelder-Mead")

    p = res.x
    return [start_point] + array_1d_to_cpts(p) + [end_point]


def _demo():
    c = bezier()
    print(c.ports["0"].y - c.ports["1"].y)
    pp.write_gds(c)
    pp.show(c)


if __name__ == "__main__":
    _demo()



""" this example shows how to add_grating couplers for single fiber in single fiber out (no fiber array)
"""

import pp

from pp.routing import route_fiber_single
from pp.add_labels import get_optical_text


def add_grating_couplers(
    component,
    get_route_factory=route_fiber_single,
    optical_io_spacing=50,
    min_input2output_spacing=200,
    optical_routing_type=2,
    bend_factory=pp.c.bend_circular,
    grating_coupler=pp.c.grating_coupler_te,
    straight_factory=pp.c.waveguide,
    with_align_ports=True,
    layer_label=pp.LAYER.LABEL,
):
    """ returns component with grating ports and labels on each port
    can add align_ports reference structure
    """
    component = pp.call_if_func(component)
    grating_coupler = pp.call_if_func(grating_coupler)

    c = pp.routing.add_fiber_array(
        component,
        optical_io_spacing=optical_io_spacing,
        bend_factory=bend_factory,
        straight_factory=straight_factory,
        grating_coupler=grating_coupler,
        get_route_factory=get_route_factory,
        optical_routing_type=optical_routing_type,
        min_input2output_spacing=min_input2output_spacing,
    )

    if with_align_ports:
        gc_port_name = list(grating_coupler.ports.keys())[0]
        gci = c << grating_coupler
        gco = c << grating_coupler
        length = c.ysize - 2 * grating_coupler.xsize
        wg = c << straight_factory(length=length)
        wg.rotate(90)
        wg.xmin = c.xmax + optical_io_spacing - grating_coupler.ysize / 2
        wg.ymin = c.ymin + grating_coupler.xsize

        gci.connect(gc_port_name, wg.ports["W0"])
        gco.connect(gc_port_name, wg.ports["E0"])

        port = wg.ports["E0"]
        label = get_optical_text(
            port, grating_coupler, 0, component_name=f"loopback_{component.name}"
        )
        c.add_label(label, position=port.midpoint, layer=layer_label)

        port = wg.ports["W0"]
        label = get_optical_text(
            port, grating_coupler, 1, component_name=f"loopback_{component.name}"
        )
        c.add_label(label, position=port.midpoint, layer=layer_label)

    return c


def test_add_grating_couplers(
    c=pp.c.ring_single_bus(gap=0.3, bend_radius=5, wg_width=0.45)
):
    cc = add_grating_couplers(c)
    assert cc
    return cc


if __name__ == "__main__":
    c = pp.c.ring_single_bus(gap=0.3, bend_radius=5, wg_width=0.45)
    cc = test_add_grating_couplers(c)
    pp.show(cc)



import inspect
import pp
from pp.components import grating_coupler_te
from pp.components import grating_coupler_tm
from pp.routing.connect import connect_strip
from pp.routing.manhattan import round_corners
from pp.add_labels import get_input_label


def connect_loop_back(port0, port1, a, b, R, y_bot_align_route):
    p0 = port0.position
    p1 = port1.position
    route = [
        p0,
        p0 + (0, a),
        p0 + (b, a),
        p0 + (b, y_bot_align_route),
        p1 + (-b, y_bot_align_route),
        p1 + (-b, a),
        p1 + (0, a),
        p1,
    ]

    bend90 = pp.c.bend_circular(radius=R)
    loop_back = round_corners(route, bend90, pp.c.waveguide)
    return loop_back


@pp.autoname
def loss_deembedding_ch13_24(
    io_sep=127.0,
    R=10.0,
    grating_coupler_function=grating_coupler_te,
    input_port_indexes=[0, 1],
):

    gc = grating_coupler_function()
    c = pp.Component()
    dx = io_sep
    gcs = [gc.ref(position=(i * dx, 0), port_id="W0", rotation=-90) for i in range(4)]

    gc_ports = [g.ports["W0"] for g in gcs]
    c.add(gcs)

    c.add(
        connect_strip(gc_ports[0], gc_ports[2], start_straight=40.0, taper_factory=None)
    )

    gsi = gc.size_info
    p1 = gc_ports[1]
    p3 = gc_ports[3]
    a = R + 5.0  # 0.5
    b = max(2 * a, io_sep / 2)
    y_bot_align_route = -gsi.width - 5.0

    c.add(connect_loop_back(p1, p3, a, b, R, y_bot_align_route))
    for i, index in enumerate(input_port_indexes):
        label = get_input_label(
            gc_ports[index], gc, i, component_name=inspect.stack()[0][3]
        )
        label.position = gc_ports[index].position
        c.add(label)

    return c


@pp.autoname
def loss_deembedding_ch12_34(
    io_sep=127.0,
    R=10.0,
    grating_coupler_function=grating_coupler_te,
    input_port_indexes=[0, 2],
):
    gc = grating_coupler_function()

    c = pp.Component()
    dx = io_sep
    gcs = [gc.ref(position=(i * dx, 0), port_id="W0", rotation=-90) for i in range(4)]

    gc_ports = [g.ports["W0"] for g in gcs]
    c.add(gcs)

    c.add(
        connect_strip(gc_ports[0], gc_ports[1], start_straight=40.0, taper_factory=None)
    )
    c.add(
        connect_strip(gc_ports[2], gc_ports[3], start_straight=40.0, taper_factory=None)
    )
    for i, index in enumerate(input_port_indexes):
        label = get_input_label(
            gc_ports[index], gc, i, component_name=inspect.stack()[0][3]
        )
        label.position = gc_ports[index].position
        c.add(label)
    return c


@pp.autoname
def loss_deembedding_ch14_23(
    io_sep=127.0,
    R=10.0,
    grating_coupler_function=grating_coupler_te,
    input_port_indexes=[0, 1],
):
    gc = grating_coupler_function()

    c = pp.Component()
    dx = io_sep
    gcs = [gc.ref(position=(i * dx, 0), port_id="W0", rotation=-90) for i in range(4)]

    gc_ports = [g.ports["W0"] for g in gcs]
    c.add(gcs)

    c.add(
        connect_strip(gc_ports[0], gc_ports[3], start_straight=40.0, taper_factory=None)
    )
    c.add(
        connect_strip(gc_ports[1], gc_ports[2], start_straight=30.0, taper_factory=None)
    )
    for i, index in enumerate(input_port_indexes):
        label = get_input_label(
            gc_ports[index], gc, i, component_name=inspect.stack()[0][3]
        )
        label.position = gc_ports[index].position
        c.add(label)
    return c


@pp.autoname
def grating_coupler_loss_te(io_sep=127.0, grating_coupler_function=grating_coupler_te):
    c = pp.Component()
    _c1 = loss_deembedding_ch13_24(grating_coupler_function=grating_coupler_function)
    _c2 = loss_deembedding_ch14_23(grating_coupler_function=grating_coupler_function)
    _c3 = loss_deembedding_ch12_34(grating_coupler_function=grating_coupler_function)
    c.add_ref(_c1)
    c2 = c.add_ref(_c2)
    c3 = c.add_ref(_c3)
    c2.movex(io_sep * 4)
    c3.movex(io_sep * 8)

    return c


@pp.autoname
def grating_coupler_loss_tm(io_sep=127.0, grating_coupler_function=grating_coupler_tm):
    c = pp.Component()
    _c1 = loss_deembedding_ch13_24(grating_coupler_function=grating_coupler_function)
    _c2 = loss_deembedding_ch14_23(grating_coupler_function=grating_coupler_function)
    _c3 = loss_deembedding_ch12_34(grating_coupler_function=grating_coupler_function)
    c.add_ref(_c1)
    c2 = c.add_ref(_c2)
    c3 = c.add_ref(_c3)
    c2.movex(io_sep * 4)
    c3.movex(io_sep * 8)

    return c


if __name__ == "__main__":
    # c = loss_deembedding_ch14_23()
    # c = loss_deembedding_ch12_34()
    # c = loss_deembedding_ch13_24()
    c = grating_coupler_loss_te()
    # c = grating_coupler_loss_tm()
    pp.show(c)



from typing import Tuple
import numpy as np
from numpy import cos, sin, pi

import pp
from pp.component import Component


@pp.autoname
def circle(
    radius: float = 10.0,
    angle_resolution: float = 2.5,
    layer: Tuple[int, int] = pp.LAYER.WG,
) -> Component:
    """ Generate a circle geometry.

    Args:
        radius: float, Radius of the circle.
        angle_resolution: float, Resolution of the curve of the ring (# of degrees per point).
        layer: (int, array-like[2], or set) Specific layer(s) to put polygon geometry on.

    .. plot::
      :include-source:

      import pp

      c = pp.c.circle(radius = 10, angle_resolution = 2.5, layer = 0)
      pp.plotgds(c)

    """

    c = pp.Component()
    t = np.linspace(0, 360, int(360 / angle_resolution) + 1) * pi / 180
    xpts = (radius * cos(t)).tolist()
    ypts = (radius * sin(t)).tolist()
    c.add_polygon(points=(xpts, ypts), layer=layer)
    return c


if __name__ == "__main__":
    c = circle()
    pp.show(c)



from phidl import device_layout as pd
import pp
from phidl.device_layout import Label
from typing import Tuple


def label(
    text: str = "abc",
    position: Tuple[float, float] = (0.0, 0.0),
    layer: Tuple[int, int] = pp.LAYER.TEXT,
) -> Label:

    gds_layer_label, gds_datatype_label = pd._parse_layer(layer)

    label_ref = pd.Label(
        text=text,
        position=position,
        anchor="o",
        layer=gds_layer_label,
        texttype=gds_datatype_label,
    )
    return label_ref


def _demo_label():
    """ there is two ways you can add labels

    """
    c = pp.Component()
    c.add_ref(pp.c.circle(radius=3.0, angle_resolution=10.0))
    # c.label(text="hi", position=(0, 1))
    c.add(label("hi", (0, 1)))
    return c


if __name__ == "__main__":

    c = _demo_label()
    pp.show(c)



""" merges multiple JSONs:

"""

import json
from omegaconf import OmegaConf
import importlib
from git import Repo
from pp.config import CONFIG, logging, get_git_hash, write_config, conf


def update_config_modules(config=conf):
    """ update config with module git hashe and version (for each module in module_requirements section)
    """
    if config.get("requirements"):
        config.update({"git_hash": get_git_hash(), "module_versions": {}})
        for module_name in config["requirements"]:
            module = importlib.import_module(module_name)
            config["module_versions"].update(
                {
                    module_name: {
                        "version": module.__version__,
                        "git_hash": Repo(module.CONFIG["repo_path"]).head.object.hexsha,
                    }
                }
            )
    return config


def merge_json(
    doe_directory=CONFIG["doe_directory"],
    extra_directories=[CONFIG["gds_directory"]],
    jsonpath=CONFIG["mask_directory"] / "metadata.json",
    json_version=6,
    config=conf,
):
    """ Merge several JSON files from config.yml
    in the root of the mask directory, gets mask_name from there

    Args:
        mask_config_directory: defaults to current working directory
        json_version:

    """
    logging.debug("Merging JSON files:")
    cells = {}
    config = config or {}
    update_config_modules(config=config)

    for directory in extra_directories + [doe_directory]:
        for filename in directory.glob("*/*.json"):
            logging.debug(filename)
            with open(filename, "r") as f:
                data = json.load(f)
                cells.update(data.get("cells"))

    does = {d.stem: json.loads(open(d).read()) for d in doe_directory.glob("*.json")}
    metadata = dict(
        json_version=json_version,
        cells=cells,
        does=does,
        config=OmegaConf.to_container(config),
    )

    write_config(metadata, jsonpath)
    print(f"Wrote  metadata in {jsonpath}")
    logging.info(f"Wrote  metadata in {jsonpath}")
    return metadata


if __name__ == "__main__":
    d = merge_json()
    print(d)

    # print(config["module_versions"])
    # pprint(d['does'])

    # with open(jsonpath, "w") as f:
    #     f.write(json.dumps(d, indent=2))



"""
lets create a crossing component with two references to other components (crossing_arm)

- add references to a component (one of the arm references rotated)
- add ports from the child references into the parent cell
- use a decorator to rename ports according to their location

"""

import pp
from pp import LAYER


@pp.autoname
def test_crossing_arm(wg_width=0.5, r1=3.0, r2=1.1, taper_width=1.2, taper_length=3.4):
    """ crossing arm
    """
    c = pp.Component()
    c << pp.c.ellipse(radii=(r1, r2), layer=LAYER.SLAB150)

    xmax = taper_length + taper_width / 2
    h = wg_width / 2
    taper_points = [
        (-xmax, h),
        (-taper_width / 2, taper_width / 2),
        (taper_width / 2, taper_width / 2),
        (xmax, h),
        (xmax, -h),
        (taper_width / 2, -taper_width / 2),
        (-taper_width / 2, -taper_width / 2),
        (-xmax, -h),
    ]

    c.add_polygon(taper_points, layer=LAYER.WG)

    c.add_port(
        name="W0", midpoint=(-xmax, 0), orientation=180, width=wg_width, layer=LAYER.WG
    )

    c.add_port(
        name="E0", midpoint=(xmax, 0), orientation=0, width=wg_width, layer=LAYER.WG
    )
    return c


@pp.port.deco_rename_ports  # This decorator will auto-rename the ports
@pp.autoname  # This decorator will generate a good name for the component
def test_crossing():
    c = pp.Component()
    arm = test_crossing_arm()

    # Create two arm references. One has a 90Deg rotation
    arm_h = arm.ref(position=(0, 0))
    arm_v = arm.ref(position=(0, 0), rotation=90)

    # Add each arm to the component
    # Also add the ports
    port_id = 0
    for a in [arm_h, arm_v]:
        c.add(a)
        for p in a.ports.values():
            # Here we don't care too much about the name we give to the ports
            # since they will be renamed. We just want the names to be unique
            c.add_port(name="{}".format(port_id), port=p)
            port_id += 1

    return c


if __name__ == "__main__":
    c = test_crossing()
    pp.show(c)



""" bends with grating couplers inside the spiral
maybe: need to add grating coupler loopback as well
"""

import numpy as np
import pp
from pp.components.bend_circular import bend_circular
from pp.components.bend_circular import bend_circular180
from pp.components import waveguide
from pp.routing import round_corners
from numpy import float64
from pp.component import Component
from typing import Callable, Optional, Tuple


def get_bend_port_distances(bend: Component) -> Tuple[float64, float64]:
    p0, p1 = bend.ports.values()
    return abs(p0.x - p1.x), abs(p0.y - p1.y)


@pp.autoname
def spiral_external_io(
    N: int = 6,
    x_inner_length_cutback: float = 300.0,
    x_inner_offset: float = 0.0,
    y_straight_inner_top: float = 0.0,
    dx: float = 3.0,
    dy: float = 3.0,
    bend90_function: Callable = bend_circular,
    bend180_function: Callable = bend_circular180,
    bend_radius: float = 50.0,
    wg_width: float = 0.5,
    straight_factory: Callable = waveguide,
    straight_factory_fall_back_no_taper: None = None,
    taper: Optional[Callable] = None,
    cutback_length: Optional[float] = None,
    **kwargs_round_corner
) -> Component:
    """

    Args:
        cutback_length: length in um, it is the approximates total length
        N: number of loops
        x_straight_inner_right:
        x_straight_inner_left:
        y_straight_inner_top:
        dx: center to center x-spacing
        dy: center to center y-spacing
        grating_coupler
        bend90_function
        bend180_function
        bend_radius
        wg_width
        straight_factory
        taper

    .. plot::
      :include-source:

      import pp

      c = pp.c.spiral_external_io()
      pp.plotgds(c)
    """
    if straight_factory_fall_back_no_taper is None:
        straight_factory_fall_back_no_taper = straight_factory

    if cutback_length:
        x_inner_length_cutback = cutback_length / (4 * (N - 1))

    y_straight_inner_top += 5

    x_inner_length_cutback += x_inner_offset
    _bend180 = pp.call_if_func(bend180_function, radius=bend_radius, width=wg_width)
    _bend90 = pp.call_if_func(bend90_function, radius=bend_radius, width=wg_width)

    rx, ry = get_bend_port_distances(_bend90)
    _, rx180 = get_bend_port_distances(_bend180)  # rx180, second arg since we rotate

    component = pp.Component()

    inner_loop_spacing = 2 * bend_radius + 5.0

    # Create manhattan path going from west grating to westest port of bend 180

    x_inner_length = x_inner_length_cutback + 5.0 + dx

    y_inner_bend = y_straight_inner_top - bend_radius - 5.0
    x_inner_loop = x_inner_length - 5.0
    p1 = (x_inner_loop, y_inner_bend)
    p2 = (x_inner_loop + inner_loop_spacing, y_inner_bend)

    _pt = np.array(p1)
    pts_w = [_pt]

    for i in range(N):
        y1 = y_straight_inner_top + ry + (2 * i + 1) * dy
        x2 = inner_loop_spacing + 2 * rx + x_inner_length + (2 * i + 1) * dx
        y3 = -ry - (2 * i + 2) * dy
        x4 = -(2 * i + 1) * dx
        if i == N - 1:
            x4 = x4 - rx180 + dx

        _pt1 = np.array([_pt[0], y1])
        _pt2 = np.array([x2, _pt1[1]])
        _pt3 = np.array([_pt2[0], y3])
        _pt4 = np.array([x4, _pt3[1]])
        _pt5 = np.array([_pt4[0], 0])
        _pt = _pt5

        pts_w += [_pt1, _pt2, _pt3, _pt4, _pt5]

    pts_w = pts_w[:-2]

    # Create manhattan path going from east grating to eastest port of bend 180
    _pt = np.array(p2)
    pts_e = [_pt]

    for i in range(N):
        y1 = y_straight_inner_top + ry + (2 * i) * dy
        x2 = inner_loop_spacing + 2 * rx + x_inner_length + 2 * i * dx
        y3 = -ry - (2 * i + 1) * dy
        x4 = -2 * i * dx

        _pt1 = np.array([_pt[0], y1])
        _pt2 = np.array([x2, _pt1[1]])
        _pt3 = np.array([_pt2[0], y3])
        _pt4 = np.array([x4, _pt3[1]])
        _pt5 = np.array([_pt4[0], 0])
        _pt = _pt5

        pts_e += [_pt1, _pt2, _pt3, _pt4, _pt5]

    pts_e = pts_e[:-2]

    # Join the two bits of paths and extrude the spiral geometry
    route_ref = round_corners(
        pts_w[::-1] + pts_e,
        bend90=_bend90,
        straight_factory=straight_factory,
        straight_factory_fall_back_no_taper=straight_factory_fall_back_no_taper,
        taper=taper,
        **kwargs_round_corner,
    )
    component.add(route_ref)
    component.absorb(route_ref)

    component.ports = route_ref.ports
    component.length = route_ref.info["length"]
    component.settings["total_length"] = route_ref.info["length"]
    component.settings["length"] = route_ref.info["length"]
    component.settings["cutback_length"] = cutback_length

    return component


if __name__ == "__main__":
    c = spiral_external_io(bend_radius=10, cutback_length=10000)
    print(c.settings["total_length"] / 1e4)
    pp.show(c)



from typing import Callable
import pp

from pp.components.mmi1x2 import mmi1x2
from pp.components.waveguide import waveguide
from pp.routing.connect import connect_strip
from pp.port import get_ports_facing


@pp.autoname
def splitter_tree(
    coupler: Callable = mmi1x2,
    n_o_outputs: int = 4,
    bend_radius: float = 10.0,
    spacing: float = 50.0,
    termination_component=None,
):
    """tree of 1x2 splitters

    Args:
        coupler: 1x2 coupler factory
        n_o_outputs:
        bend_radius: for routing
        spacing: 2X spacing
        termination_component: factory or component for EAST termination

    .. code::

             __|
          __|  |__
        _|  |__
         |__       spacing


    .. plot::
      :include-source:

      import pp

      c = pp.c.splitter_tree(coupler=pp.c.mmi1x2(), n_o_outputs=4, spacing=50, bend_radius=10)
      pp.plotgds(c)

    """
    n_o_outputs = n_o_outputs
    c = pp.Component()

    coupler = pp.call_if_func(coupler)
    _coupler = c.add_ref(coupler)
    coupler_sep = coupler.ports["E1"].y - coupler.ports["E0"].y

    if n_o_outputs > 2:
        _cmp = splitter_tree(
            coupler=coupler,
            n_o_outputs=n_o_outputs // 2,
            bend_radius=bend_radius,
            spacing=spacing / 2,
        )
    else:
        termination_component = (
            termination_component
            if termination_component is not None
            else waveguide(length=0.1)
        )
        _cmp = pp.call_if_func(termination_component)

    spacing = (
        spacing
        if spacing is not None
        else _cmp.ports["W0"].y - _coupler.size_info.south
    )
    if spacing < coupler_sep:
        tree_top = _cmp.ref(port_id="W0", position=_coupler.ports["E1"])
        tree_bot = _cmp.ref(
            port_id="W0", position=_coupler.ports["E0"], v_mirror=False  # True
        )

    else:
        d = 2 * bend_radius + 1
        spacing = max(spacing, d)

        tree_top = _cmp.ref(
            port_id="W0", position=_coupler.ports["E1"].position + (d, spacing)
        )
        tree_bot = _cmp.ref(
            port_id="W0",
            position=_coupler.ports["E0"].position + (d, -spacing),
            v_mirror=False,  # True,
        )

        c.add(connect_strip(coupler.ports["E1"], tree_top.ports["W0"]))
        c.add(connect_strip(coupler.ports["E0"], tree_bot.ports["W0"]))

    i = 0
    for p in get_ports_facing(tree_bot, "E"):
        c.add_port(name="{}".format(i), port=p)
        i += 1

    for p in get_ports_facing(tree_top, "E"):
        c.add_port(name="{}".format(i), port=p)
        i += 1

    c.add(tree_bot)
    c.add(tree_top)
    c.add_port(name="W0", port=_coupler.ports["W0"])

    return c


if __name__ == "__main__":
    c = splitter_tree(coupler=pp.c.mmi1x2())
    pp.show(c)



import pytest
import pp
from pp.components import component_type2factory

component_types = [
    "waveguide",
    "bend_circular",
    # "bend_euler90",
    "coupler",
    "mmi1x2",
    "mmi2x2",
]


@pytest.mark.parametrize("component_type", component_types)
def test_sp_load(component_type, data_regression):
    c = component_type2factory[component_type]()
    sp = pp.sp.load(c)

    port_names = sp[0]
    f = list(sp[1])
    s = sp[2]

    lenf = s.shape[0]
    rows = s.shape[1]
    cols = s.shape[2]

    assert rows == cols == len(c.ports)
    assert len(port_names) == len(c.ports)
    data_regression.check(dict(port_names=port_names))
    assert lenf == len(f)


if __name__ == "__main__":
    c = pp.c.waveguide(layer=(2, 0))
    print(c.get_sparameters_path())



from typing import Callable, Tuple
from pp.component import Component
from pp.layers import LAYER
from pp.name import autoname
from pp.components.waveguide import waveguide


@autoname
def waveguide_pin(
    length: float = 10.0,
    width: float = 0.5,
    width_i: float = 0.0,
    width_p: float = 1.0,
    width_n: float = 1.0,
    width_pp: float = 1.0,
    width_np: float = 1.0,
    width_ppp: float = 1.0,
    width_npp: float = 1.0,
    layer_p: Tuple[int, int] = LAYER.P,
    layer_n: Tuple[int, int] = LAYER.N,
    layer_pp: Tuple[int, int] = LAYER.Pp,
    layer_np: Tuple[int, int] = LAYER.Np,
    layer_ppp: Tuple[int, int] = LAYER.Ppp,
    layer_npp: Tuple[int, int] = LAYER.Npp,
    waveguide_factory: Callable = waveguide,
) -> Component:
    """PN doped waveguide

        .. code::


                               |<------width------>|
                                ____________________
                               |     |       |     |
            ___________________|     |       |     |__________________________|
                                     |       |                                |
                P++     P+     P     |   I   |     N        N+         N++    |
            __________________________________________________________________|
                                                                              |
                                     |width_i| width_n | width_np | width_npp |
                                        0    oi        on        onp         onpp

    .. plot::
      :include-source:

      import pp

      c = pp.c.waveguide_pin(length=10)
      pp.plotgds(c)

    """
    c = waveguide_factory(length=length, width=width)

    oi = width_i / 2
    on = oi + width_n
    onp = oi + width_n + width_np
    onpp = oi + width_n + width_np + width_npp

    # N doping
    c.add_polygon([(0, oi), (length, oi), (length, onpp), (0, onpp)], layer=layer_n)

    if layer_np:
        c.add_polygon(
            [(0, on), (length, on), (length, onpp), (0, onpp)], layer=layer_np
        )
    if layer_npp:
        c.add_polygon(
            [(0, onp), (length, onp), (length, onpp), (0, onpp)], layer=layer_npp
        )

    oi = -width_i / 2
    op = oi - width_p
    opp = oi - width_p - width_pp
    oppp = oi - width_p - width_pp - width_ppp

    # P doping
    c.add_polygon([(0, oi), (length, oi), (length, oppp), (0, oppp)], layer=layer_p)
    if layer_pp:
        c.add_polygon(
            [(0, op), (length, op), (length, oppp), (0, oppp)], layer=layer_pp
        )
    if layer_ppp:
        c.add_polygon(
            [(0, opp), (length, opp), (length, oppp), (0, oppp)], layer=layer_ppp
        )

    return c


if __name__ == "__main__":
    import pp

    c = waveguide_pin(width_i=1, pins=True)
    pp.show(c)



import pp
from pp.port import rename_ports_by_orientation
from pp.components.bend_circular import bend_circular
from pp.components.coupler_ring import coupler_ring
from pp.components.waveguide import waveguide
from pp.drc import assert_on_2nm_grid
from pp.component import Component
from pp.config import call_if_func
from pp.name import autoname
from pp.components.taper import taper


@autoname
def ring_single_dut(
    component,
    wg_width=0.5,
    gap=0.2,
    length_x=4,
    bend_radius=5,
    length_y=0,
    coupler=coupler_ring,
    waveguide=waveguide,
    bend=bend_circular,
    with_dut=True,
):
    """ single bus ring made of two couplers (ct: top, cb: bottom)
    connected with two vertical waveguides (wyl: left, wyr: right)
    DUT (Device Under Test) in the middle to extract loss from quality factor


    Args:
        with_dut: if False changes dut for just a waveguide

    .. code::

          bl-wt-br
          |      | length_y
          wl     dut
          |      |
         --==cb==-- gap

          length_x
    """
    dut = pp.call_if_func(component)
    dut = rename_ports_by_orientation(dut)

    assert_on_2nm_grid(gap)

    coupler = call_if_func(
        coupler, gap=gap, wg_width=wg_width, bend_radius=bend_radius, length_x=length_x
    )
    waveguide_side = call_if_func(
        waveguide, width=wg_width, length=length_y + dut.xsize
    )
    waveguide_top = call_if_func(waveguide, width=wg_width, length=length_x)
    bend = call_if_func(bend, width=wg_width, radius=bend_radius)

    c = Component()
    cb = c << coupler
    wl = c << waveguide_side
    if with_dut:
        d = c << dut
    else:
        d = c << waveguide_side
    bl = c << bend
    br = c << bend
    wt = c << waveguide_top

    wl.connect(port="E0", destination=cb.ports["N0"])
    bl.connect(port="N0", destination=wl.ports["W0"])

    wt.connect(port="W0", destination=bl.ports["W0"])
    br.connect(port="N0", destination=wt.ports["E0"])
    d.connect(port="W0", destination=br.ports["W0"])
    c.add_port("E0", port=cb.ports["E0"])
    c.add_port("W0", port=cb.ports["W0"])
    return c


if __name__ == "__main__":
    c = ring_single_dut(component=taper(width2=3), pins=True)
    pp.show(c)



import pp
from pp.components.waveguide import waveguide
from pp.components.bend_circular import bend_circular
from pp.name import autoname
from pp.component import Component
from typing import Callable


@autoname
def coupler90(
    bend_radius: float = 10.0,
    width: float = 0.5,
    gap: float = 0.2,
    waveguide_factory: Callable = waveguide,
    bend90_factory: Callable = bend_circular,
) -> Component:
    """ Waveguide coupled to a bend with gap

    Args:
        bend_radius: um
        width: waveguide width (um)
        gap: um

    .. plot::
      :include-source:

      import pp
      c = pp.c.coupler90()
      pp.plotgds(c)

    """
    # pp.drc.assert_on_1nm_grid((width + gap) / 2)
    y = pp.drc.snap_to_1nm_grid((width + gap) / 2)

    c = Component()

    wg = c << waveguide_factory(length=bend_radius, width=width, pins=False)
    bend = c << bend90_factory(radius=bend_radius, width=width, pins=False)

    pbw = bend.ports["W0"]
    bend.movey(pbw.midpoint[1] + gap + width)

    # This component is a leaf cell => using absorb
    c.absorb(wg)
    c.absorb(bend)

    port_width = 2 * width + gap

    c.add_port(port=wg.ports["E0"], name="E0")
    c.add_port(port=bend.ports["N0"], name="N0")
    c.add_port(name="W0", midpoint=[0, y], width=port_width, orientation=180)
    return c


if __name__ == "__main__":
    c = coupler90(width=0.45, gap=0.3)
    pp.show(c)
    # print(c.ports)
